import numpy as np
import pandas as pd
import os
import torch.utils.data as data_utils
import torch
import math
import scanpy as sc
from sklearn.model_selection import train_test_split
from dance.datasets.multimodality import JointEmbeddingNIPSDataset
from dance.transforms.preprocess import calculate_log_library_size
import argparse
import anndata as ad
import dance.utils.metrics as metrics
from dance.modules.multi_modality.joint_embedding.dcca import DCCA

# from scMVAE.utilities import read_dataset, normalize, calculate_log_library_size, parameter_setting, save_checkpoint, \
#     load_checkpoint, adjust_learning_rate
from dance.modules.multi_modality.joint_embedding.scmvae import scMVAE

def parameter_setting():
    parser = argparse.ArgumentParser(description='Single cell Multi-omics data analysis')

    outPath = './new_test/'

    # parser.add_argument('--File1', '-F1', type=str, default='5-counts-RNA.tsv', help='input file name1')
    # parser.add_argument('--File2', '-F2', type=str, default='0.75-0.2-counts-ATAC.tsv', help='input file name2')
    # parser.add_argument('--File2_1', '-F2_1', type=str, default='0.75-0.2-counts-ATAC_binary.tsv',
    #                     help='input file name2_1')
    #
    # parser.add_argument('--File3', '-F3', type=str, default='5-cellinfo-RNA.tsv', help='input meta file')
    # parser.add_argument('--File_combine', '-F_com', type=str, default='Gene_chromatin_order_combine.tsv',
    #                     help='input combine file name')
    # parser.add_argument('--File_mofa', '-F_mofa', type=str, default='MOFA_combine_cluster.csv',
    #                     help='cluster for mofa predicted')

    parser.add_argument ('--latent_fusion', '-olf1', type=str, default='First_simulate_fusion.csv',
                        help='fusion latent code file')
    parser.add_argument('--latent_1', '-ol1', type=str, default='scRNA_latent_combine.csv',
                        help='first latent code file')
    parser.add_argument('--latent_2', '-ol2', type=str, default='scATAC_latent.csv', help='seconde latent code file')
    parser.add_argument('--denoised_1', '-od1', type=str, default='scRNA_seq_denoised.csv',
                        help='outfile for denoised file1')
    parser.add_argument('--normalized_1', '-on1', type=str, default='scRNA_seq_normalized_combine.tsv',
                        help='outfile for normalized file1')
    parser.add_argument('--denoised_2', '-od2', type=str, default='scATAC_seq_denoised.csv',
                        help='outfile for denoised file2')

    parser.add_argument('--workdir', '-wk', type=str, default=outPath, help='work path')
    parser.add_argument('--outdir', '-od', type=str, default=outPath, help='Output path')

    parser.add_argument('--lr', type=float, default=1E-3, help='Learning rate')
    parser.add_argument('--weight_decay', type=float, default=1e-6, help='weight decay')
    parser.add_argument('--eps', type=float, default=0.01, help='eps')

    parser.add_argument('--batch_size', '-b', type=int, default=64, help='Batch size')

    parser.add_argument('--seed', type=int, default=200, help='Random seed for repeat results')
    parser.add_argument('--latent', '-l', type=int, default=10, help='latent layer dim')
    parser.add_argument('--max_epoch', '-me', type=int, default=20, help='Max epoches')
    parser.add_argument('--max_iteration', '-mi', type=int, default=3000, help='Max iteration')
    parser.add_argument('--anneal_epoch', '-ae', type=int, default=200, help='Anneal epoch')
    parser.add_argument('--epoch_per_test', '-ept', type=int, default=5, help='Epoch per test')
    parser.add_argument('--max_ARI', '-ma', type=int, default=-200, help='initial ARI')
    parser.add_argument('-t', '--subtask', default='openproblems_bmmc_cite_phase2')
    parser.add_argument('-device', '--device', default='cuda:3')
    parser.add_argument('--final_rate', type=float, default = 1e-4)
    parser.add_argument('--scale_factor', type=float, default=4)

    return parser

if __name__ == "__main__":
    parser = parameter_setting()
    args = parser.parse_args()

    args.sf1 = 5
    args.sf2 = 1
    args.cluster1 = args.cluster2 = 4
    args.lr1 = 0.01
    args.flr1 = 0.001
    args.lr2 = 0.005
    args.flr2 = 0.0005

    dataset = JointEmbeddingNIPSDataset(args.subtask, data_dir='./data/joint_embedding').load_data() \
        .load_metadata().load_sol().preprocess('feature_selection')

    adata = dataset.modalities[0]
    adata1 = dataset.modalities[1]
    idx = np.random.permutation(adata.shape[0])
    train_index = idx[:int(len(idx) * 0.85)]
    test_index = idx[int(len(idx) * 0.85):]
    label_ground_truth = adata.obs['batch']

    Nfeature1 = np.shape(adata.X)[1]
    Nfeature2 = np.shape(adata1.X)[1]

    device = torch.device(args.device)

    model = DCCA( layer_e_1 = [Nfeature1, 128], hidden1_1 = 128, Zdim_1 = 4, layer_d_1 = [4, 128],
                   hidden2_1 = 128, layer_e_2 = [Nfeature2, 1500, 128], hidden1_2 = 128, Zdim_2 = 4,
                   layer_d_2 = [4], hidden2_2 = 4, args = args, ground_truth1 = label_ground_truth,
                   ground_truth2 = label_ground_truth, Type_1 = "NB", Type_2 = "Bernoulli", cycle = 1,
                   attention_loss = "Eucli" ).to(device)

    model.to(device)
    train = data_utils.TensorDataset(torch.from_numpy(adata[train_index].X.todense()),
                                     torch.from_numpy(adata.layers['counts'][train_index].todense()),
                                     torch.from_numpy(adata.obs['size_factors'][train_index].values),
                                     torch.from_numpy(adata1[train_index].X.todense()),
                                     torch.from_numpy(adata1.layers['counts'][train_index].todense()),
                                     torch.from_numpy(adata.obs['size_factors'][train_index].values))

    train_loader = data_utils.DataLoader(train, batch_size=args.batch_size, shuffle=True)

    test = data_utils.TensorDataset(torch.from_numpy(adata[test_index].X.todense()),
                                    torch.from_numpy(adata.layers['counts'][test_index].todense()),
                                    torch.from_numpy(adata.obs['size_factors'][test_index].values),
                                    torch.from_numpy(adata1[test_index].X.todense()),
                                    torch.from_numpy(adata1.layers['counts'][test_index].todense()),
                                    torch.from_numpy(adata.obs['size_factors'][test_index].values))

    test_loader = data_utils.DataLoader(test, batch_size=args.batch_size, shuffle=False)

    total = data_utils.TensorDataset(torch.from_numpy(adata.X.todense()),
                                     torch.from_numpy(adata.layers['counts'].todense()),
                                     torch.from_numpy(adata.obs['size_factors'].values),
                                     torch.from_numpy(adata1.X.todense()),
                                     torch.from_numpy(adata1.layers['counts'].todense()),
                                     torch.from_numpy(adata.obs['size_factors'].values))
    total_loader = data_utils.DataLoader(total, batch_size=args.batch_size, shuffle=False)

    model.fit(train_loader, test_loader, total_loader, "RNA")

    with torch.no_grad():
        emb1, emb2 = model.predict(total_loader)

    embeds = np.concatenate([emb1, emb2], 1)
    print(embeds)
    print(model.score(total_loader))

    mod1_obs = dataset.modalities[0].obs
    mod1_uns = dataset.modalities[0].uns
    adata = ad.AnnData(
        X=embeds,
        obs=mod1_obs,
        uns={
            'dataset_id': mod1_uns['dataset_id'],
            'method_id': 'scmogcn',
        },
    )
    # adata.write_h5ad(f'./joint_embedding_scmvae.h5ad', compression="gzip")

    metrics.labeled_clustering_evaluate(adata, dataset)


import sys
import os.path as o
#use if running from dance/examples/spatial/cell_type_deconvo
root_path=o.abspath(o.join(o.dirname(sys.modules[__name__].__file__), "../../.."))
sys.path.append(root_path)


from dance.datasets.spatial import CellTypeDeconvoDataset
from dance.transforms.preprocess import pseudo_spatial_process, preprocess_adj, split
from dance.transforms.graph_construct import stAdjConstruct
import torch
import torch.nn.functional as F
import numpy as np
import scanpy as sc
import random
from dance.modules.spatial.cell_type_deconvo.dstg import DSTGLearner
# Set random seed
seed = 123
np.random.seed(seed)
device = torch.device("cuda" if torch.cuda.is_available() else "cpu")


import warnings
warnings.filterwarnings("ignore")


# get data 
dataset = CellTypeDeconvoDataset("mouse brain 1", data_dir="data/spatial")
## dataset.data has repeat name , be careful

count_data = [sc.AnnData(dataset.data['ref_sc_count']), sc.AnnData(dataset.data['mix_count'])]
label_data = [dataset.data['ref_sc_count_annot'], dataset.data['mix_count_annot']]
#pre-process: get variable genes --> normalize --> log1p --> standardize --> out
#set scRNA to false if already using pseudo spot data with real spot data

st_data, labels, hvgs = pseudo_spatial_process(count_data,label_data, clust_vr='subclass', scRNA=True, n_hvg=250, N_p=500)

labels = [lab.drop(['cell_count', 'total_umi_count', 'n_counts'], axis=1) for lab in labels]
# create train/val/test split
adj_data, features, labels_binary_train, labels_binary_val, labels_binary_test, train_mask, pred_mask, val_mask, test_mask, new_label, true_label = split(st_data, labels, pre_process=1, split_val=.8)


labels_binary_train=torch.FloatTensor(labels_binary_train)

features = torch.sparse.FloatTensor(torch.LongTensor([features[0][:,0].tolist(), features[0][:,1].tolist()]),
                                  torch.FloatTensor(features[1]))

#construct adjacency matrix
adj = stAdjConstruct(st_data, labels, adj_data, k_filter=50)

#preprocess adjacency matrix
adj = preprocess_adj(adj)

adj = torch.sparse.FloatTensor(torch.LongTensor([adj.row.tolist(), adj.col.tolist()]),
                              torch.FloatTensor(adj.data.astype(np.int32)))

# Create model
model_func = DSTGLearner
model = model_func(nfeat=features.size()[1], nhid1=32, nout=labels_binary_train.size()[1],device=device, bias=False, dropout=0,act=F.relu)

#fit model
model.fit(features, adj, labels_binary_train, labels_mask=torch.FloatTensor(train_mask), lr=0.01, max_epochs=10, weight_decay=0)

#predict on both pseudo & real ST data
model.predict(features, adj)



# -*- coding: utf-8 -*-
import os
import random
import argparse
from argparse import Namespace

import dance.transforms.preprocess
from dance.transforms.graph_construct import construct_modality_prediction_graph, gen_batch_features
from dance.modules.multi_modality.predict_modality.scmogcn import *
from dance.datasets.multimodality import ModalityPredictionDataset
from dance.utils import set_seed

def pipeline(transductive = True, verbose=2, logger=None, **kwargs):
    PREFIX = kwargs['prefix']
    os.makedirs(kwargs["log_folder"], exist_ok=True)
    os.makedirs(kwargs["model_folder"], exist_ok=True)
    os.makedirs(kwargs["result_folder"], exist_ok=True)
    if verbose>1:
        logger = open(f'{kwargs["log_folder"]}/{PREFIX}.log', 'w')
        logger.write(str(kwargs) + '\n')

    subtask = kwargs['subtask']

    dataset = ModalityPredictionDataset(subtask).load_data()
    dataset.download_pathway()
    if kwargs['preprocessing']!='none':
        dataset = dataset.preprocess(kwargs['preprocessing'])

    idx = np.random.permutation(dataset.modalities[0].shape[0])
    split = {'train': idx[:-int(len(idx) * 0.15)], 'valid': idx[-int(len(idx) * 0.15):]}

    input_train_mod1 = dataset.sparse_features()[0]
    input_train_mod2 = dataset.sparse_features()[1]
    if transductive:
        input_test_mod1 = dataset.sparse_features()[2]
        true_test_mod2 = dataset.sparse_features()[3]

    FEATURE_SIZE = input_train_mod1.shape[1]
    CELL_SIZE = input_train_mod1.shape[0] + input_test_mod1.shape[0] if transductive else input_train_mod1.shape[0]
    OUTPUT_SIZE = input_train_mod2.shape[1]
    TRAIN_SIZE = input_train_mod1.shape[0]

    kwargs['FEATURE_SIZE'] = FEATURE_SIZE
    kwargs['CELL_SIZE'] = CELL_SIZE
    kwargs['OUTPUT_SIZE'] = OUTPUT_SIZE
    kwargs['TRAIN_SIZE'] = TRAIN_SIZE

    if kwargs['inductive'] != 'trans':
        g, gtest = construct_modality_prediction_graph(dataset, **kwargs)
    else:
        gtest = g = construct_modality_prediction_graph(dataset, **kwargs)

    if not kwargs['no_batch_features']:
        if transductive:
            batch_features = gen_batch_features([dataset.modalities[0], dataset.modalities[2]])
        else:
            batch_features = gen_batch_features(dataset.modalities[:1])
        kwargs['BATCH_NUM'] = batch_features.shape[1]
        if kwargs['inductive'] != 'trans':
            g.nodes['cell'].data['bf'] = batch_features[:TRAIN_SIZE]
            gtest.nodes['cell'].data['bf'] = batch_features
        else:
            g.nodes['cell'].data['bf'] = batch_features

    device = kwargs['device']
    g = g.to(device)
    if kwargs['inductive'] != 'trans':
        gtest = gtest.to(device)

    # data loader
    y = torch.from_numpy(input_train_mod2.toarray()).to(device)
    if transductive:
        y_test = torch.from_numpy(true_test_mod2.toarray()).to(device)

    model = ScMoGCNWrapper(Namespace(**kwargs))

    if kwargs['sampling']:
        model.fit_with_sampling(g, y, split, transductive, verbose, y_test, logger)
    else:
        model.fit(g, y, split, transductive, verbose, y_test, logger)

    print(model.predict(g, np.arange(TRAIN_SIZE, CELL_SIZE)))
    print(model.score(g, np.arange(TRAIN_SIZE, CELL_SIZE), y_test))

if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('-prefix', '--prefix', default='dance_openproblems_bmmc_atac2rna_test')
    parser.add_argument('-t', '--subtask', default='openproblems_bmmc_multiome_phase2_mod2')
    parser.add_argument('-pww', '--pathway_weight', default='pearson', choices=['cos', 'one', 'pearson'])
    parser.add_argument('-pwth', '--pathway_threshold', type=float, default=-1.0)
    parser.add_argument('-l', '--log_folder', default='./logs')
    parser.add_argument('-m', '--model_folder', default='./models')
    parser.add_argument('-r', '--result_folder', default='./results')
    parser.add_argument('-e', '--epoch', type=int, default=10000)
    parser.add_argument('-nbf', '--no_batch_features', action='store_true')
    parser.add_argument('-npw', '--no_pathway', action='store_true')
    parser.add_argument('-opw', '--only_pathway', action='store_true')
    parser.add_argument('-res', '--residual', default='res_cat', choices=['none', 'res_add', 'res_cat'])
    parser.add_argument('-inres', '--initial_residual', action='store_true')
    parser.add_argument('-pwagg', '--pathway_aggregation', default='alpha',
                        choices=['sum', 'attention', 'two_gate', 'one_gate', 'alpha', 'cat'])
    parser.add_argument('-pwalpha', '--pathway_alpha', type=float, default=0.5)
    parser.add_argument('-nrc', '--no_readout_concatenate', action='store_true')
    parser.add_argument('-bs', '--batch_size', default=1000, type=int)
    parser.add_argument('-nm', '--normalization', default='group', choices=['batch', 'layer', 'group', 'none'])
    parser.add_argument('-ac', '--activation', default='gelu', choices=['leaky_relu', 'relu', 'prelu', 'gelu'])
    parser.add_argument('-em', '--embedding_layers', default=1, type=int, choices=[1, 2, 3])
    parser.add_argument('-ro', '--readout_layers', default=1, type=int, choices=[1, 2])
    parser.add_argument('-conv', '--conv_layers', default=4, type=int, choices=[1, 2, 3, 4, 5, 6])
    parser.add_argument('-agg', '--agg_function', default='mean', choices=['gcn', 'mean'])
    parser.add_argument('-device', '--device', default='cuda:3')
    parser.add_argument('-sb', '--save_best', action='store_true')
    parser.add_argument('-sf', '--save_final', action='store_true')
    parser.add_argument('-lr', '--learning_rate', type=float, default=1e-2)
    parser.add_argument('-lrd', '--lr_decay', type=float, default=0.99)
    parser.add_argument('-wd', '--weight_decay', type=float, default=1e-5)
    parser.add_argument('-hid', '--hidden_size', type=int, default=48)
    parser.add_argument('-edd', '--edge_dropout', type=float, default=0.3)
    parser.add_argument('-mdd', '--model_dropout', type=float, default=0.2)
    parser.add_argument('-es', '--early_stopping', type=int, default=0)
    parser.add_argument('-c', '--cpu', type=int, default=1)
    parser.add_argument('-or', '--output_relu', default='none', choices=['relu', 'leaky_relu', 'none'])
    parser.add_argument('-i', '--inductive', default='trans', choices=['normal', 'opt', 'trans'])
    parser.add_argument('-sa', '--subpath_activation', action='store_true')
    parser.add_argument('-ci', '--cell_init', default='none', choices=['none', 'pca'])
    parser.add_argument('-bas', '--batch_seperation', action='store_true')
    parser.add_argument('-pwpath', '--pathway_path', default='./data/h.all.v7.4')
    parser.add_argument('-seed', '--random_seed', type=int, default=777)#random.randint(0, 2147483647))
    parser.add_argument('-ws', '--weighted_sum', action='store_true')
    parser.add_argument('-samp', '--sampling', action='store_true')
    parser.add_argument('-ns', '--node_sampling_rate', type=float, default=0.5)
    parser.add_argument('-prep', '--preprocessing', default='none', choices=['none', 'feature_selection', 'pca'])

    args = parser.parse_args()

    # For test only (low gpu memory setting; to reproduce competition result need >20G memory - v100)
    if True:
        args.preprocessing = 'feature_selection'
        args.no_pathway = True
        args.sampling = True
        args.batch_size = 10000
        args.epoch = 10
    elif args.subtask == 'openproblems_bmmc_multiome_phase2_mod2':
        args.preprocessing = 'feature_selection'
    elif args.subtask in ['openproblems_bmmc_cite_phase2_mod2', 'openproblems_bmmc_multiome_phase2_rna']:
        args.sampling = True
        args.edge_dropout = 0

    # Regular settings
    if args.subtask.find('rna') == -1:
        args.no_pathway = True
    if args.sampling:
        args.no_pathway = True

    set_seed(args.random_seed)
    torch.set_num_threads(args.cpu)

    pipeline(**vars(args))




import numpy as np
import pandas as pd
import os
import torch.utils.data as data_utils
import torch
import math
import scanpy as sc
from sklearn.model_selection import train_test_split
from dance.datasets.multimodality import JointEmbeddingNIPSDataset
from dance.transforms.preprocess import calculate_log_library_size
import argparse
import anndata as ad
import dance.utils.metrics as metrics

# from scMVAE.utilities import read_dataset, normalize, calculate_log_library_size, parameter_setting, save_checkpoint, \
#     load_checkpoint, adjust_learning_rate
from dance.modules.multi_modality.joint_embedding.scmvae import scMVAE

def parameter_setting():
    parser = argparse.ArgumentParser(description='Single cell Multi-omics data analysis')

    outPath = './new_test/'

    # parser.add_argument ('--latent_fusion', '-olf1', type=str, default='First_simulate_fusion.csv',
    #                     help='fusion latent code file')
    # parser.add_argument('--latent_1', '-ol1', type=str, default='scRNA_latent_combine.csv',
    #                     help='first latent code file')
    # parser.add_argument('--latent_2', '-ol2', type=str, default='scATAC_latent.csv', help='seconde latent code file')
    # parser.add_argument('--denoised_1', '-od1', type=str, default='scRNA_seq_denoised.csv',
    #                     help='outfile for denoised file1')
    # parser.add_argument('--normalized_1', '-on1', type=str, default='scRNA_seq_normalized_combine.tsv',
    #                     help='outfile for normalized file1')
    # parser.add_argument('--denoised_2', '-od2', type=str, default='scATAC_seq_denoised.csv',
    #                     help='outfile for denoised file2')

    parser.add_argument('--workdir', '-wk', type=str, default=outPath, help='work path')
    parser.add_argument('--outdir', '-od', type=str, default=outPath, help='Output path')

    parser.add_argument('--lr', type=float, default=1E-3, help='Learning rate')
    parser.add_argument('--weight_decay', type=float, default=1e-6, help='weight decay')
    parser.add_argument('--eps', type=float, default=0.01, help='eps')

    parser.add_argument('--batch_size', '-b', type=int, default=64, help='Batch size')
    # parser.add_argument('--use_cuda', dest='use_cuda', default=True, action='store_true',
    #                     help=" whether use cuda(default: True)")

    parser.add_argument('--seed', type=int, default=200, help='Random seed for repeat results')
    parser.add_argument('--latent', '-l', type=int, default=10, help='latent layer dim')
    parser.add_argument('--max_epoch', '-me', type=int, default=20, help='Max epoches')
    parser.add_argument('--max_iteration', '-mi', type=int, default=3000, help='Max iteration')
    parser.add_argument('--anneal_epoch', '-ae', type=int, default=200, help='Anneal epoch')
    parser.add_argument('--epoch_per_test', '-ept', type=int, default=5, help='Epoch per test, must smaller than max iteration.')
    parser.add_argument('--max_ARI', '-ma', type=int, default=-200, help='initial ARI')
    parser.add_argument('-t', '--subtask', default='openproblems_bmmc_cite_phase2')
    parser.add_argument('-device', '--device', default='cuda:3')
    parser.add_argument('--final_rate', type=float, default = 1e-4)
    parser.add_argument('--scale_factor', type=float, default=4)

    return parser

if __name__ == "__main__":
    parser = parameter_setting()
    args = parser.parse_args()
    assert args.max_iteration>args.epoch_per_test

    dataset = JointEmbeddingNIPSDataset(args.subtask, data_dir='./data/joint_embedding').load_data() \
        .load_metadata().load_sol().preprocess(kind='feature_selection')

    adata1 = dataset.modalities[0]
    adata2 = dataset.modalities[1]
    adata1.X = adata1.layers['counts']
    adata2.X = adata2.layers['counts']

    idx = np.random.permutation(adata1.shape[0])
    train_index = idx[:int(len(idx) * 0.85)]
    valid_index = idx[int(len(idx) * 0.85):]

    lib_mean1, lib_var1 = calculate_log_library_size(adata1.X)
    lib_mean2, lib_var2 = calculate_log_library_size(adata2.X)

    Nfeature1 = np.shape(adata1.X)[1]
    Nfeature2 = np.shape(adata2.X)[1]

    device = torch.device(args.device)

    model = scMVAE(encoder_1=[Nfeature1, 1024, 128, 128],
                   hidden_1=128,
                   Z_DIMS=22,
                   decoder_share=[22, 128, 256],
                   share_hidden=128,
                   decoder_1=[128, 128, 1024],
                   hidden_2=1024,
                   encoder_l=[Nfeature1, 128],
                   hidden3=128,
                   encoder_2=[Nfeature2, 1024, 128, 128],
                   hidden_4=128,
                   encoder_l1=[Nfeature2, 128],
                   hidden3_1=128,
                   decoder_2=[128, 128, 1024],
                   hidden_5=1024,
                   drop_rate=0.1,
                   log_variational=True,
                   Type="ZINB",
                   device=device,
                   n_centroids=22,
                   penality="GMM",
                   model=1, )

    args.lr = 0.001
    args.anneal_epoch = 200

    model.to(device)
    train = data_utils.TensorDataset(torch.from_numpy(adata1.X[train_index].todense()),
                                     torch.from_numpy(lib_mean1[train_index]),
                                     torch.from_numpy(lib_var1[train_index]),
                                     torch.from_numpy(lib_mean2[train_index]),
                                     torch.from_numpy(lib_var2[train_index]),
                                     torch.from_numpy(adata2.X[train_index].todense()))

    valid = data_utils.TensorDataset(torch.from_numpy(adata1.X[valid_index].todense()),
                                    torch.from_numpy(lib_mean1[valid_index]),
                                    torch.from_numpy(lib_var1[valid_index]),
                                    torch.from_numpy(lib_mean2[valid_index]),
                                    torch.from_numpy(lib_var2[valid_index]),
                                    torch.from_numpy(adata2.X[valid_index].todense()))

    total = data_utils.TensorDataset(torch.from_numpy(adata1.X.todense()),
                                     torch.from_numpy(adata2.X.todense()))

    total_loader = data_utils.DataLoader(total, batch_size=args.batch_size, shuffle=False)
    model.init_gmm_params(total_loader)
    model.fit(args, train, valid, args.final_rate, args.scale_factor, device)

    # model.load_state_dict(torch.load('./saved_model/model_best.pth.tar') )

    test_mod1 = dataset.tensor_features(0).to(device)
    test_mod2 = dataset.tensor_features(1).to(device)
    embeds = model.predict(test_mod1, test_mod2).cpu().numpy()
    print(embeds)
    print(model.score(test_mod1, test_mod2, dataset.test_sol))

    # mod1_obs = dataset.modalities[0].obs
    # mod1_uns = dataset.modalities[0].uns
    # adata = ad.AnnData(
    #     X=embeds,
    #     obs=mod1_obs,
    #     uns={
    #         'dataset_id': mod1_uns['dataset_id'],
    #         'method_id': 'scmogcn',
    #     },
    # )
    # # adata.write_h5ad(f'./joint_embedding_scmvae.h5ad', compression="gzip")
    #
    # metrics.labeled_clustering_evaluate(adata, dataset)


import argparse
import glob
import os
import os.path as osp
import sys
from dataclasses import dataclass

import anndata as ad
import h5py
import numpy as np
import pandas as pd
import scanpy as sc
import torch
from torch.utils.data import Dataset

from dance.data import download_file, download_unzip
from dance.transforms.preprocess import (get_map_dict, load_actinn_data, load_annotation_data_internal,
                                         load_annotation_test_data, load_imputation_data_internal, load_svm_data,
                                         prepare_data_celltypist, splitCommonAnnData)


@dataclass
class CellTypeDatasetParams:
    random_seed = None
    dense_dim = None
    proj_path = None
    train_dataset = None
    test_dataset = None
    species = None
    tissue = None
    score = None
    gpu = None
    evaluate = None
    train_dir = None
    test_dir = None
    statistics_path = None
    map_path = None
    filetype = None
    threshold = None
    exclude_rate = None
    test_rate = None


class CellTypeDataset():
    def __init__(self, data_type="scdeepsort", random_seed=10, proj_path="./", dense_dim=None, train_dataset=None,
                 test_dataset=None, species=None, tissue=None, score=True, gpu=None, evaluate=None, train_dir=None,
                 test_dir=None, statistics_path=None, map_path=None, filetype=None, threshold=None, exclude_rate=None,
                 test_rate=None, data_dir="./", train_set=None, train_label=None, test_set=None, test_label=None,
                 X_celltypist=None, labels_celltypist=None, genes_celltypist=None, transpose_input_celltypist=False,
                 singlecellnet_type="Lung", svm_settings=None):
        self.data_dir = data_dir
        self.data_type = data_type
        if data_type == "scdeepsort":
            self.download_pretrained_data()
        self.params = CellTypeDatasetParams()
        self.params.random_seed = random_seed
        self.params.dense_dim = dense_dim
        self.params.proj_path = proj_path
        self.params.train_dataset = train_dataset
        self.params.test_dataset = test_dataset
        self.params.species = species
        self.params.tissue = tissue
        self.params.score = score
        self.params.gpu = gpu
        self.params.evaluate = evaluate
        self.params.train_dir = train_dir
        self.params.test_dir = test_dir
        self.params.statistics_path = statistics_path
        self.params.map_path = map_path
        self.params.filetype = filetype
        self.params.threshold = threshold
        self.params.exclude_rate = exclude_rate
        self.params.test_rate = test_rate
        self.train_set = train_set
        self.train_label = train_label
        self.test_set = test_set
        self.test_label = test_label
        self.X_celltypist = X_celltypist
        self.labels_celltypist = labels_celltypist
        self.genes_celltypist = genes_celltypist
        self.transpose_input_celltypist = transpose_input_celltypist
        self.singlecellnet_type = singlecellnet_type

    def download_all_data(self):
        # download data
        gene_class = ["human_cell_atlas", "human_test_data", "mouse_cell_atlas", "mouse_test_data"]

        url = {
            "human_cell_atlas": "https://www.dropbox.com/s/1itq1pokplbqxhx/human_cell_atlas.zip?dl=0",
            "human_test_data": "https://www.dropbox.com/s/gpxjnnvwyblv3xb/human_test_data.zip?dl=0",
            "mouse_cell_atlas": "https://www.dropbox.com/s/ng8d3eujfah9ppl/mouse_cell_atlas.zip?dl=0",
            "mouse_test_data": "https://www.dropbox.com/s/pkr28czk5g3al2p/mouse_test_data.zip?dl=0"
        }

        file_name = {
            "human_cell_atlas": "human_cell_atlas.zip?dl=0",
            "human_test_data": "human_test_data.zip?dl=0",
            "mouse_cell_atlas": "mouse_cell_atlas.zip?dl=0",
            "mouse_test_data": "mouse_test_data.zip?dl=0"
        }

        species = {
            "human_cell_atlas": "human",
            "human_test_data": "human",
            "mouse_cell_atlas": "mouse",
            "mouse_test_data": "mouse"
        }

        os.system("mkdir " + self.data_dir + "/train")
        os.system("mkdir " + self.data_dir + "/train/mouse")
        os.system("mkdir " + self.data_dir + "/train/human")
        for class_name in gene_class:
            os.system("wget " + url[class_name])
            os.system("unzip " + file_name[class_name])
            os.system("mv " + class_name + "/* " + self.data_dir + "/train/" + species[class_name] + "/")
            os.system("rm " + file_name[class_name])
            os.system("rm -r " + class_name)

        os.system("cp -r " + self.data_dir + "/train/ " + self.data_dir + "/test")

    def download_benchmark_data(self, download_map=True, download_pretrained=True):

        if self.is_benchmark_complete():
            return

        urls = {
            "train_mouse_Brain753_celltype": "https://www.dropbox.com/s/x2katwk93z06sgw/mouse_Brain753_celltype.csv?dl=1",
            "train_mouse_Brain753_data": "https://www.dropbox.com/s/3f3wbplgo3xa4ww/mouse_Brain753_data.csv?dl=1",
            "train_mouse_Brain3285_celltype": "https://www.dropbox.com/s/ozsobozk3ihkrqg/mouse_Brain3285_celltype.csv?dl=1",
            "train_mouse_Brain3285_data": "https://www.dropbox.com/s/zjrloejx8iqdqsa/mouse_Brain3285_data.csv?dl=1",
            "train_mouse_Kidney4682_celltype": "https://www.dropbox.com/s/3plrve7g9v428ec/mouse_Kidney4682_celltype.csv?dl=1",
            "train_mouse_Kidney4682_data": "https://www.dropbox.com/s/olf5nirtieu1ikq/mouse_Kidney4682_data.csv?dl=1",
            "train_mouse_Spleen1970_celltype": "https://www.dropbox.com/s/3ea64vk546fjxvr/mouse_Spleen1970_celltype.csv?dl=1",
            "train_mouse_Spleen1970_data": "https://www.dropbox.com/s/c4te0fr1qicqki8/mouse_Spleen1970_data.csv?dl=1",
            "test_mouse_Brain2695_celltype": "https://www.dropbox.com/s/gh72dk7i0p7fggu/mouse_Brain2695_celltype.csv?dl=1",
            "test_mouse_Brain2695_data": "https://www.dropbox.com/s/ufianih66xjqxdu/mouse_Brain2695_data.csv?dl=1",
            "test_mouse_Kidney203_celltype": "https://www.dropbox.com/s/t4eyaig889qdiz2/mouse_Kidney203_celltype.csv?dl=1",
            "test_mouse_Kidney203_data": "https://www.dropbox.com/s/kmos1ceubumgmpj/mouse_Kidney203_data.csv?dl=1",
            "test_mouse_Spleen1759_celltype": "https://www.dropbox.com/s/gczehvgai873mhb/mouse_Spleen1759_celltype.csv?dl=1",
            "test_mouse_Spleen1759_data": "https://www.dropbox.com/s/fl8t7rbo5dmznvq/mouse_Spleen1759_data.csv?dl=1",
        }

        # Download training and testing data
        for name, url in urls.items():
            filename = url.split("/")[-1].split("?")[0]
            filepath = osp.join(self.data_dir, *name.split("_")[:2], filename)
            download_file(url, filepath)

        if download_map:
            # Download mapping data
            download_unzip("https://www.dropbox.com/sh/hw1189sgm0kfrts/AAAapYOblLApqygZ-lGo_70-a?dl=1",
                           osp.join(self.data_dir, "map"))

        if download_pretrained:
            # Download pretrained stats data
            download_unzip("https://www.dropbox.com/sh/s2cxcrzl2ama9zp/AACKwiYtS8hbOOudQLIMDvXUa?dl=1",
                           osp.join(self.data_dir, "pretrained"))

    def download_example_data(self):
        # download example data
        os.system("wget https://www.dropbox.com/s/5lbdytwpip0kl58/example.zip?dl=0")
        os.system(f"mv example.zip?dl=0 {self.data_dir}")
        os.system(f"unzip {self.data_dir} /example.zip?dl=0")
        os.system(f"mv {self.data_dir}/example/* {self.data_dir}")
        os.system(f"rm {self.data_dir}/example.zip?dl=0")
        os.system(f"rm -r {self.data_dir}/example")

    def download_pretrained_data(self):
        """Download pretrained model and label maps."""
        os.system("wget https://www.dropbox.com/s/i5mufqwc1hy97m4/pretrained.zip?dl=0")
        os.system(f"mv pretrained.zip?dl=0 {self.data_dir}")
        os.system(f"unzip {self.data_dir}/pretrained.zip?dl=0")
        os.system(f"rm {self.data_dir}/pretrained.zip?dl=0")

    def download_actinn_data(self):
        """Download pretrained model and label maps"""
        os.system("wget https://www.dropbox.com/s/me2zu2rok9twcjg/actinn_data.zip?dl=0")
        os.system("unzip actinn_data.zip?dl=0")
        os.system("rm actinn_data.zip?dl=0")

    def download_singlecellnet_data(self):
        """Download pretrained model and label maps"""
        print("Downloading Training Data from Bladder...")
        os.system("wget https://cnobjects.s3.amazonaws.com/singleCellNet/pySCN/training/adBladder_TabSen_101320.h5ad")
        print("Downloading Training Data from Fat...")
        os.system("wget https://cnobjects.s3.amazonaws.com/singleCellNet/pySCN/training/adFat_TabSen_101320.h5ad")
        print("Downloading Training Data from Heart...")
        os.system("wget https://cnobjects.s3.amazonaws.com/singleCellNet/pySCN/training/adHeart_TabSen_101320.h5ad")
        print("Downloading Training Data from Kidney...")
        os.system("wget https://cnobjects.s3.amazonaws.com/singleCellNet/pySCN/training/adKidney_TabSen_101320.h5ad")
        print("Downloading Training Data from Large Intestine...")
        os.system("wget https://cnobjects.s3.amazonaws.com/singleCellNet/pySCN/training/adL_Intestine_TabSen_101320.h5ad")
        print("Downloading Training Data from Lung...")
        os.system("wget https://cnobjects.s3.amazonaws.com/singleCellNet/pySCN/training/adLung_TabSen_100920.h5ad")
        print("Downloading Training Data from Mammary Gland...")
        os.system("wget https://cnobjects.s3.amazonaws.com/singleCellNet/pySCN/training/adMammary_Gland_TabSen_101320.h5ad")
        print("Downloading Training Data from Marrow...")
        os.system("wget https://cnobjects.s3.amazonaws.com/singleCellNet/pySCN/training/adMarrow_TabSen_101320.h5ad")
        print("Downloading Training Data from Pancreas...")
        os.system("wget https://cnobjects.s3.amazonaws.com/singleCellNet/pySCN/training/adPancreas_TabSen_101320.h5ad")
        print("Downloading Training Data from Skeletal Muscle...")
        os.system("wget https://cnobjects.s3.amazonaws.com/singleCellNet/pySCN/training/adSkel_Muscle_TabSen_101320.h5ad")
        print("Downloading Training Data from Skin...")
        os.system("wget https://cnobjects.s3.amazonaws.com/singleCellNet/pySCN/training/adSkin_TabSen_101320.h5ad")
        print("Downloading Training Data from Trachea...")
        os.system("wget https://cnobjects.s3.amazonaws.com/singleCellNet/pySCN/training/adTrachea_TabSen_101320.h5ad")

    def download_singlecellnet_lung_example_data(self):
        print("Downloading lung example data...")
        os.system("wget https://cnobjects.s3.amazonaws.com/singleCellNet/pySCN/query/GSE124872_raw_counts_single_cell.mtx.gz")
        os.system("gzip -d  GSE124872_raw_counts_single_cell.mtx.gz")
        os.system("wget https://cnobjects.s3.amazonaws.com/singleCellNet/pySCN/query/GSE124872_Angelidis_2018_metadata.csv")
        os.system("wget https://cnobjects.s3.amazonaws.com/singleCellNet/pySCN/query/genes.csv")

    def is_complete(self):
        """Check if data is complete."""
        check = [
            osp.join(self.data_dir, "train"),
            osp.join(self.data_dir, "test"),
            osp.join(self.data_dir, "pretrained")
        ]

        for i in check:
            if not os.path.exists(i):
                print(f"file {i} doesn't exist")
                return False
        return True

    def is_benchmark_complete(self):
        check = [
            "test_mouse_Brain2695_celltype.csv",
            "test_mouse_Brain2695_data.csv",
            "test_mouse_Kidney203_celltype.csv",
            "test_mouse_Kidney203_data.csv",
            "test_mouse_Spleen1759_celltype.csv",
            "test_mouse_Spleen1759_data.csv",
            "train_mouse_Brain3285_celltype.csv",
            "train_mouse_Brain3285_data.csv",
            "train_mouse_Brain753_celltype.csv",
            "train_mouse_Brain753_data.csv",
            "train_mouse_Kidney4682_celltype.csv",
            "train_mouse_Kidney4682_data.csv",
            "train_mouse_Spleen1970_celltype.csv",
            "train_mouse_Spleen1970_data.csv",
        ]
        for name in check:
            filename = name[name.find('mouse'):]
            file_i = osp.join(self.data_dir, *name.split("_")[:2], filename)
            if not os.path.exists(file_i):
                print(file_i)
                print(f"file {filename} doesn't exist")
                return False
        # check maps
        map_check = [
            osp.join(self.data_dir, "map", "mouse", "map.xlsx"),
            osp.join(self.data_dir, "map", "human", "map.xlsx"),
            osp.join(self.data_dir, "map", "celltype2subtype.xlsx")
        ]
        for file in map_check:
            if not os.path.exists(file):
                print(f"file {name} doesn't exist")
                return False
        # TODO: check pretrained data
        return True

    def is_actinn_complete(self):
        # check whether data is complete or not
        check = ["test_data"]
        for i in check:
            if not os.path.exists(i):
                print(f"file {i} doesn't exist")
                return False
        return True

    def is_singlecellnet_complete(self):
        # check whether data is complete or not
        check = [
            "Bladder",
            "Fat",
            "Heart",
            "Kidney",
            "L_Intestine",
            "Lung",
            "Mammary_Gland",
            "Marrow",
            "Pancreas",
            "Skel_Muscle",
            "Skin",
            "Trachea",
        ]
        for i in check:
            if i == "Lung":
                file_i = "adLung_TabSen_100920.h5ad"
            else:
                file_i = "ad" + i + "_TabSen_101320.h5ad"
            if not os.path.exists(file_i):
                print(f"file {i} doesn't exist")
                return False
        return True

    def load_data(self):
        # Load data from existing h5ad files, or download files and load data.
        if self.data_type == "scdeepsort" or self.data_type == "scdeepsort_exp":
            if self.is_complete():
                pass
            else:
                if self.data_type == "scdeepsort":
                    self.download_all_data()
                if self.data_type == "scdeepsort_exp":
                    self.download_benchmark_data()
                assert self.is_complete()

            (
                self.num_cells,
                self.num_genes,
                self.num_labels,
                self.graph,
                self.train_ids,
                self.test_ids,
                self.labels,
            ) = load_annotation_data_internal(self.params)

            if self.params.score:
                (
                    self.total_cell_test,
                    self.num_genes_test,
                    self.num_labels_test,
                    self.id2label_test,
                    self.test_dict,
                    self.test_label_dict,
                    self.time_used_test,
                ) = load_annotation_test_data(self.params)
            else:
                (
                    self.total_cell_test,
                    self.num_genes_test,
                    self.num_labels_test,
                    self.id2label_test,
                    self.test_dict,
                    self.time_used_test,
                ) = load_annotation_test_data(self.params)

        if self.data_type == "svm":
            if self.is_complete():
                pass
            else:
                if self.data_type == "svm":
                    self.download_benchmark_data()
                if self.data_type == "svm_exp":
                    self.download_example_data()
                assert self.is_complete()

            (
                self.svm_num_labels,
                self.svm_train_labels,
                self.svm_train_cell_feat,
                self.svm_map_dict,
                self.svm_id2label,
                self.svm_test_label_dict,
                self.svm_test_feat_dict,
                self.svm_test_cell_id_dict,
            ) = load_svm_data(self.params)

        if self.data_type == "actinn":
            self.download_benchmark_data()
            (
                self.train_set,
                self.train_label,
                self.test_set,
                self.test_label,
                self.barcode,
                self.label_to_type_dict,
            ) = load_actinn_data(self.train_set, self.train_label, self.test_set, self.test_label)

        if self.data_type == "celltypist":
            self.download_benchmark_data(download_pretrained=False)
            self.map_dict = get_map_dict(self.params.map_path, self.params.tissue)  # load map
            train_data = pd.read_csv(
                osp.join(self.params.proj_path, self.params.train_dir, self.params.species,
                         self.params.species + "_" + self.params.tissue + str(self.params.train_dataset) + "_data.csv"),
                index_col=0)
            train_celltype = pd.read_csv(
                osp.join(
                    self.params.proj_path, self.params.train_dir, self.params.species,
                    self.params.species + "_" + self.params.tissue + str(self.params.train_dataset) + "_celltype.csv"),
                index_col=1)
            self.train_adata = ad.AnnData(train_data.T, train_celltype)
            test_data = pd.read_csv(
                osp.join(self.params.proj_path, self.params.test_dir, self.params.species,
                         self.params.species + "_" + self.params.tissue + str(self.params.test_dataset) + "_data.csv"),
                index_col=0)
            test_celltype = pd.read_csv(
                osp.join(
                    self.params.proj_path, self.params.test_dir, self.params.species,
                    self.params.species + "_" + self.params.tissue + str(self.params.test_dataset) + "_celltype.csv"),
                index_col=1)
            self.test_adata = ad.AnnData(test_data.T, test_celltype)

            if self.X_celltypist is not None:
                self.indata, self.labels, self.genes = prepare_data_celltypist(self.X_celltypist,
                                                                               self.labels_celltypist,
                                                                               self.genes_celltypist,
                                                                               self.transpose_input_celltypist)
            else:
                adata_James = sc.read("celltypist_demo_folder/gut_cell_atlas_James.h5ad",
                                      backup_url="https://cellgeni.cog.sanger.ac.uk/gutcellatlas/Colon_cell_atlas.h5ad")
                adata_Elmentaite = sc.read(
                    "celltypist_demo_folder/gut_cell_atlas_Elmentaite.h5ad",
                    backup_url="https://cellgeni.cog.sanger.ac.uk/gutcellatlas/Full_obj_raw_counts_nosoupx.h5ad")
                sc.pp.normalize_total(adata_Elmentaite, target_sum=1e4)
                sc.pp.log1p(adata_Elmentaite)
                self.adata_James = adata_James
                self.adata_Elmentaite = adata_Elmentaite
                self.indata, self.labels, self.genes = prepare_data_celltypist(adata_James, "cell_type", None, False)

        if self.data_type == "singlecellnet_exp":
            if self.is_singlecellnet_complete():
                pass
            else:
                self.download_singlecellnet_data()
                self.download_singlecellnet_lung_example_data()

            if self.singlecellnet_type == "Lung":
                file_i = "adLung_TabSen_100920.h5ad"
            else:
                file_i = "ad" + self.singlecellnet_type + "_TabSen_101320.h5ad"
            adTrain = sc.read(file_i)
            qDatT = sc.read_mtx("GSE124872_raw_counts_single_cell.mtx")
            qDat = qDatT.T
            genes = pd.read_csv("genes.csv")
            qDat.var_names = genes.x
            qMeta = pd.read_csv("GSE124872_Angelidis_2018_metadata.csv")
            qMeta.columns.values[0] = "cellid"
            qMeta.index = qMeta["cellid"]
            qDat.obs = qMeta.copy()
            genesTrain = adTrain.var_names
            genesQuery = qDat.var_names
            cgenes = genesTrain.intersection(genesQuery)
            adTrain1 = adTrain[:, cgenes]
            adQuery = qDat[:, cgenes].copy()
            adQuery = adQuery[adQuery.obs["nGene"] >= 500, :].copy()
            self.expTrain, self.expVal = splitCommonAnnData(adTrain1, ncells=200, dLevel="cell_ontology_class")

        if self.data_type == "singlecellnet":
            self.download_benchmark_data(download_pretrained=False)
            self.map_dict = get_map_dict(self.params.map_path, self.params.tissue)  # load map
            train_data = pd.read_csv(
                osp.join(self.params.proj_path, self.params.train_dir, self.params.species,
                         self.params.species + "_" + self.params.tissue + str(self.params.train_dataset) + "_data.csv"),
                index_col=0)
            train_celltype = pd.read_csv(
                osp.join(
                    self.params.proj_path, self.params.train_dir, self.params.species,
                    self.params.species + "_" + self.params.tissue + str(self.params.train_dataset) + "_celltype.csv"),
                index_col=1)
            self.train_adata = ad.AnnData(train_data.T, train_celltype)
            test_data = pd.read_csv(
                osp.join(self.params.proj_path, self.params.test_dir, self.params.species,
                         self.params.species + "_" + self.params.tissue + str(self.params.test_dataset) + "_data.csv"),
                index_col=0)
            test_celltype = pd.read_csv(
                osp.join(
                    self.params.proj_path, self.params.test_dir, self.params.species,
                    self.params.species + "_" + self.params.tissue + str(self.params.test_dataset) + "_celltype.csv"),
                index_col=1)
            self.test_adata = ad.AnnData(test_data.T, test_celltype)

            if self.is_singlecellnet_complete():
                pass
            else:
                self.download_singlecellnet_data()
                self.download_singlecellnet_lung_example_data()

            if self.singlecellnet_type == "Lung":
                file_i = "adLung_TabSen_100920.h5ad"
            else:
                file_i = "ad" + self.singlecellnet_type + "_TabSen_101320.h5ad"
            adTrain = sc.read(file_i)
            qDatT = sc.read_mtx("GSE124872_raw_counts_single_cell.mtx")
            qDat = qDatT.T
            genes = pd.read_csv("genes.csv")
            qDat.var_names = genes.x
            qMeta = pd.read_csv("GSE124872_Angelidis_2018_metadata.csv")
            qMeta.columns.values[0] = "cellid"
            qMeta.index = qMeta["cellid"]
            qDat.obs = qMeta.copy()
            genesTrain = adTrain.var_names
            genesQuery = qDat.var_names
            cgenes = genesTrain.intersection(genesQuery)
            adTrain1 = adTrain[:, cgenes]
            adQuery = qDat[:, cgenes].copy()
            adQuery = adQuery[adQuery.obs["nGene"] >= 500, :].copy()
            self.expTrain, self.expVal = splitCommonAnnData(adTrain1, ncells=200, dLevel="cell_ontology_class")

        return self


class ClusteringDataset():
    """
    Data downloading and loading for clustering

    Args:
            data_dir (str, optional): Path to store datasets. Defaults to "./data".
            dataset (str, optional): Choice of dataset. Defaults to "mouse_bladder_cell",
                                     choices=['10X_PBMC', 'mouse_bladder_cell', 'mouse_ES_cell', 'worm_neuron_cell'].
    """

    def __init__(self, data_dir="./data", dataset="mouse_bladder_cell"):
        self.data_dir = data_dir
        self.dataset = dataset
        self.X = None
        self.Y = None

    def download_data(self):
        # download data
        data_url = {
            "10X_PBMC": "https://www.dropbox.com/s/pfunm27qzgfpj3u/10X_PBMC.h5?dl=1",
            "mouse_bladder_cell": "https://www.dropbox.com/s/xxtnomx5zrifdwi/mouse_bladder_cell.h5?dl=1",
            "mouse_ES_cell": "https://www.dropbox.com/s/zbuku7oznvji8jk/mouse_ES_cell.h5?dl=1",
            "worm_neuron_cell": "https://www.dropbox.com/s/58fkgemi2gcnp2k/worm_neuron_cell.h5?dl=1"
        }
        data_name = self.dataset
        download_file(data_url.get(data_name), self.data_dir + "/{}.h5".format(data_name))
        return self

    def is_complete(self):
        # judge data is complete or not
        return os.path.exists(os.path.join(self.data_dir, f"{self.dataset}.h5"))

    def load_data(self):
        # Load data from existing h5ad files, or download files and load data.
        if self.is_complete():
            pass
        else:
            self.download_data()
            assert self.is_complete()

        data_mat = h5py.File(f"{self.data_dir}/{self.dataset}.h5", "r")
        self.X = np.array(data_mat["X"])
        self.Y = np.array(data_mat["Y"])
        return self


class PretrainDataset(Dataset):
    """Dataset object for scDSC pretraining."""

    def __init__(self, data):
        self.x = data

    def __len__(self):
        return self.x.shape[0]

    def __getitem__(self, idx):
        return torch.from_numpy(np.array(self.x[idx])).float(), \
               torch.from_numpy(np.array(idx))


class TrainingDataset(Dataset):
    """Dataset object for scDSC training."""

    def __init__(self, X, Y):
        self.x = X
        self.y = Y

    def __len__(self):
        return self.x.shape[0]

    def __getitem__(self, idx):
        return torch.from_numpy(np.array(self.x[idx])), \
               torch.from_numpy(np.array(self.y[idx])), \
               torch.from_numpy(np.array(idx))


@dataclass
class ImputationDatasetParams:
    data_dir = None
    random_seed = None
    min_counts = None
    train_dataset = None
    test_dataset = None
    gpu = None
    filetype = None


class ImputationDatasetParams:
    data_dir = None
    random_seed = None
    min_counts = None
    train_dataset = None
    test_dataset = None
    gpu = None
    filetype = None


class ImputationDataset():
    def __init__(
            self,
            random_seed = 10,
            gpu = -1,
            # evaluate = None,
            filetype = None,
            data_dir = "data",
            train_dataset = "human_stemcell",
            test_dataset = "pbmc",
            min_counts = 1):
        parser = argparse.ArgumentParser()
        self.params = parser.parse_args()
        self.params.data_dir = data_dir
        self.params.random_seed = random_seed
        self.params.min_counts = min_counts
        self.params.train_dataset = train_dataset
        self.params.test_dataset = test_dataset
        self.params.gpu = gpu
        # self.params.evaluate = evaluate
        self.params.filetype = filetype
        # self.download_all_data()
        # self.download_pretrained_data()

    def download_all_data(self):
        # download data

        gene_class = ["pbmc_data", "mouse_brain_data", "mouse_embryo_data", "human_stemcell_data"]

        url = {"pbmc_data": "https://www.dropbox.com/s/brj3orsjbhnhawa/5k.zip?dl=0",
        "mouse_embryo_data": "https://www.dropbox.com/s/8ftx1bydoy7kn6p/GSE65525.zip?dl=0",
        "mouse_brain_data": "https://www.dropbox.com/s/zzpotaayy2i29hk/neuron_10k.zip?dl=0",
        "human_stemcell_data": "https://www.dropbox.com/s/g2qua2j3rqcngn6/GSE75748.zip?dl=0"}

        file_name = {"pbmc_data": "5k.zip?dl=0",
        "mouse_embryo_data": "GSE65525.zip?dl=0",
        "mouse_brain_data": "neuron_10k.zip?dl=0",
        "human_stemcell_data": "GSE75748.zip?dl=0"}

        dl_files ={"pbmc_data": "5k_*",
        "mouse_embryo_data": "GSE65525",
        "mouse_brain_data": "neuron*",
        "human_stemcell_data": "GSE75748"
        }

        dataset_to_file = {"pbmc_data": "5k_pbmc_protein_v3_filtered_feature_bc_matrix.h5",
        "mouse_embryo_data": list(map(lambda x: "GSE65525/" + x, ["GSM1599494_ES_d0_main.csv",
                                                                  "GSM1599497_ES_d2_LIFminus.csv",
                                                                  "GSM1599498_ES_d4_LIFminus.csv",
                                                                  "GSM1599499_ES_d7_LIFminus.csv"])),
        "mouse_brain_data": "neuron_10k_v3_filtered_feature_bc_matrix.h5",
        "human_stemcell_data": "GSE75748/GSE75748_sc_time_course_ec.csv.gz"}
        self.params.dataset_to_file = dataset_to_file
        if sys.platform != 'win32':
            if not os.path.exists(self.params.data_dir):
                os.system("mkdir " + self.params.data_dir)
            if not os.path.exists(self.params.data_dir + "/train"):
                os.system("mkdir " + self.params.data_dir + "/train")

            for class_name in gene_class:
                if not any(list(map(os.path.exists, glob.glob(self.params.data_dir + "/train/" + class_name + "/" + dl_files[class_name])))):
                    os.system("mkdir " + self.params.data_dir + "/train/" + class_name)
                    # if not any(list(map(os.path.exists, glob.glob("./" + dl_files[class_name])))):
                    os.system("wget " + url[class_name])  # assumes linux... mac needs to install
                    os.system("unzip " + file_name[class_name])
                    os.system("rm " + file_name[class_name])
                    os.system("mv " + dl_files[class_name] +" " + self.params.data_dir + "/train/" + class_name + "/")
                # if class_name == 'pbmc_data':
                #     os.system("mv " + "./5k_* " + self.params.data_dir + "/train/" + class_name + "/")
                # elif class_name == 'mouse_embryo_data':
                #     os.system("mv " + "./GSE65525 " + self.params.data_dir + "/train/" + class_name + "/")
                # elif class_name == 'mouse_brain_data':
                #     os.system("mv " + "./neuron* " + self.params.data_dir + "/train/" + class_name + "/")
                # elif class_name == 'human_stemcell_data':
                #     os.system("mv " + "./GSE75748 " + self.params.data_dir + "/train/" + class_name + "/")
            os.system("cp -r " + self.params.data_dir + "/train/ " + self.params.data_dir + "/test")
        if sys.platform == 'win32':
            if not os.path.exists(self.params.data_dir):
                os.system("mkdir " + self.params.data_dir)
            if not os.path.exists(self.params.data_dir + "/train"):
                # os.system('chdir ' + self.params.data_dir)
                # os.system("mkdir /train")
                # os.system('chdir ../')
                # os.system("mkdir " + self.params.data_dir + "/train")
                os.mkdir(self.params.data_dir + "/train")
            for class_name in gene_class:
                if not any(list(map(os.path.exists, glob.glob(
                        self.params.data_dir + "/train/" + class_name + "/" + dl_files[class_name])))):
                    # os.chdir(self.params.data_dir)
                    # os.system("mkdir /train/")
                    # os.chdir('train')
                    # os.system("mkdir " + class_name)
                    os.mkdir(self.params.data_dir + "/train/" + class_name)
                    # os.system("mkdir " + self.params.data_dir + "/train/" + class_name)
                    # os.chdir('../..')
                    # if not any(list(map(os.path.exists, glob.glob("./" + dl_files[class_name])))):
                    os.system("curl " + url[class_name])
                    os.system("tar -xf " + file_name[class_name])
                    # os.system("rd /s /q " + file_name[class_name])
                    os.system("del -R " + file_name[class_name])
                    os.system("move " + dl_files[class_name] + " " + self.params.data_dir + "/train/" + class_name + "/")
                # if class_name == 'pbmc_data':
                #     os.system("mv " + "./5k_* " + self.params.data_dir + "/train/" + class_name + "/")
                # elif class_name == 'mouse_embryo_data':
                #     os.system("mv " + "./GSE65525 " + self.params.data_dir + "/train/" + class_name + "/")
                # elif class_name == 'mouse_brain_data':
                #     os.system("mv " + "./neuron* " + self.params.data_dir + "/train/" + class_name + "/")
                # elif class_name == 'human_stemcell_data':
                #     os.system("mv " + "./GSE75748 " + self.params.data_dir + "/train/" + class_name + "/")
            os.system("copy /r " + self.params.data_dir + "/train/ " + self.params.data_dir + "/test")

#    def download_example_data(self):
#        # download example data
#        os.system("wget " + "https://www.dropbox.com/s/lcl0m9ph4fcki08/example.zip?dl=0")
#        os.system("mv " + "example.zip?dl=0 " + self.params.data_dir)
#        os.system("unzip " + self.params.data_dir + "/example.zip?dl=0")
#        os.system("rm " + self.params.data_dir + "/example.zip?dl=1")

#    def download_pretrained_data(self):
#        #download pretrained model
#        os.system("wget " + "https://www.dropbox.com/s/i5mufqwc1hy97m4/pretrained.zip?dl=0")
#        os.system("mv " + "pretrained.zip?dl=0 " + self.params.data_dir)
#        os.system("unzip " + self.params.data_dir + "pretrained.zip?dl=0")
#        os.system("rm " + self.params.data_dir + "pretrained.zip?dl=0")

    def is_complete(self):
        # check whether data is complete or not
        check = [
            self.params.data_dir + "/train",
            self.params.data_dir + "/test",
        ]
        # self.params.data_dir + "/pretrained"]

        for i in check:
            if not os.path.exists(i):
                print("file {} doesn't exist".format(i))
                return False
        return True

    def load_data(self, model_params, model='GraphSCI'):
        # Load data from existing h5ad files, or download files and load data.
        if self.is_complete():
            pass
        else:
            self.download_all_data()
            assert self.is_complete()
        # self.params.dataset_to_file = {"pbmc_data": "5k_pbmc_protein_v3_raw_feature_bc_matrix.h5",
        #                                "mouse_embryo_data": list(
        #                                    map(lambda x: "GSE65525/" + x, ["GSM1599494_ES_d0_main.csv",
        #                                                                    "GSM1599497_ES_d2_LIFminus.csv",
        #                                                                    "GSM1599498_ES_d4_LIFminus.csv",
        #                                                                    "GSM1599499_ES_d7_LIFminus.csv"])),
        #                                "mouse_brain_data": "neuron_10k_v3_raw_feature_bc_matrix.h5",
        #                                "human_stemcell_data": "GSE75748/GSE75748_sc_time_course_ec.csv.gz"}

        data_dict = load_imputation_data_internal(self.params, model_params, model=model)
        self.params.num_cells = data_dict['num_cells']
        self.params.num_genes = data_dict['num_genes']
        self.params.train_data = data_dict['train_data']
        self.params.test_data = data_dict['test_data']
        self.params.adata = data_dict['adata']

        if model == 'GraphSCI':
            self.params.train_data_raw = data_dict['train_data_raw']
            self.params.test_data_raw = data_dict['test_data_raw']
            self.params.adj_train = data_dict['adj_train']
            self.params.adj_test = data_dict['adj_test']
            self.params.adj_train_false = data_dict['adj_train_false']
            self.params.adj_norm_train = data_dict['adj_norm_train']
            self.params.adj_norm_test = data_dict['adj_norm_test']
            self.params.size_factors = data_dict['size_factors']
            self.params.train_size_factors = data_dict['train_size_factors']
            self.params.test_size_factors = data_dict['test_size_factors']
            self.params.train_size_factors = data_dict['train_size_factors']
            self.params.test_size_factors = data_dict['test_size_factors']
            self.params.test_idx = data_dict['test_idx']
        if model == 'DeepImpute':
            self.params.X_train = self.params.train_data[0]
            self.params.Y_train = self.params.train_data[1]
            self.params.X_test = self.params.test_data[0]
            self.params.Y_test = self.params.test_data[1]
            self.params.inputgenes = data_dict['predictors']
            self.params.targetgenes = data_dict['targets']
            self.params.total_counts = data_dict['total_counts']
            self.params.true_counts = data_dict['true_counts']
            self.params.genes_to_impute = data_dict['genes_to_impute']
        if model == 'scGNN':
            self.params.genelist = data_dict['genelist']
            self.params.celllist = data_dict['celllist']
            self.params.test_idx = data_dict['test_idx']
        """
        total_cell, num_genes, num_labels, id2label, test_dict, time_used = load_imputation_test_data(self.params)
        self.total_cell_test = total_cell
        self.num_genes_test = num_genes
        self.num_labels_test = num_labels
        self.id2label_test = id2label
        self.test_dict_test = test_dict
        self.time_used_test = time_used
        """

        return self



"""
This is the implementation of SpaGCN.
Extended from https://github.com/jianhuupenn/SpaGCN

Reference
----------
Hu, Jian, et al. "SpaGCN: Integrating gene expression, spatial location and histology to identify spatial domains and spatially variable genes by graph convolutional network." Nature methods 18.11 (2021): 1342-1351.
"""

from sklearn.cluster import KMeans
import torch
import torch.nn as nn 
import torch.nn.functional as F
import numpy as np 
import pandas as pd 
# from models import GraphConvolution
from torch.nn.parameter import Parameter
import torch.optim as optim
from dance import utils
import math
from sklearn.decomposition import PCA
import scanpy as sc
from scipy.sparse import issparse
import random

def refine(sample_id, pred, dis, shape="hexagon"):
    """
    An optional refinement step for the clustering result.
    In this step, SpaGCN examines the domain assignment of 
    each spot and its surrounding spots. For a given spot, 
    if more than half of its surrounding spots are assigned 
    to a different domain, this spot will be relabeled to 
    the same domain as the major label of its surrounding spots.
    
    Parameters
    ----------
    sample_id : 
        sample id
    pred : 
        initial prediction
    dis : 
        graph structure
    shape : str optional
        by default as "hexagon"
    
    
    Returns
    -------
    refined_pred : 
        refined prediction.
    
    """
    refined_pred=[]
    pred=pd.DataFrame({"pred": pred}, index=sample_id)
    dis_df=pd.DataFrame(dis, index=sample_id, columns=sample_id)
    if shape=="hexagon":
        num_nbs=6 
    elif shape=="square":
        num_nbs=4
    else:
        print("Shape not recongized, shape='hexagon' for Visium data, 'square' for ST data.")
    for i in range(len(sample_id)):
        index=sample_id[i]
        dis_tmp=dis_df.loc[index, :].sort_values()
        nbs=dis_tmp[0:num_nbs+1]
        nbs_pred=pred.loc[nbs.index, "pred"]
        self_pred=pred.loc[index, "pred"]
        v_c=nbs_pred.value_counts()
        if (v_c.loc[self_pred]<num_nbs/2) and (np.max(v_c)>num_nbs/2):
            refined_pred.append(v_c.idxmax())
        else:           
            refined_pred.append(self_pred)
    return refined_pred
    
class GraphConvolution(nn.Module):
    """
    Simple GCN layer, similar to https://arxiv.org/abs/1609.02907
    """
    def __init__(self, in_features, out_features, bias=True):
        super(GraphConvolution, self).__init__()
        self.in_features = in_features
        self.out_features = out_features
        self.weight = Parameter(torch.FloatTensor(in_features, out_features))
        if bias:
            self.bias = Parameter(torch.FloatTensor(out_features))
        else:
            self.register_parameter('bias', None)
        self.reset_parameters()

    def reset_parameters(self):
        stdv = 1. / math.sqrt(self.weight.size(1))
        self.weight.data.uniform_(-stdv, stdv)
        if self.bias is not None:
            self.bias.data.uniform_(-stdv, stdv)

    def forward(self, input, adj):
        support = torch.mm(input, self.weight)
        output = torch.spmm(adj, support)
        if self.bias is not None:
            return output + self.bias
        else:
            return output

    def __repr__(self):
        return self.__class__.__name__ + ' (' \
               + str(self.in_features) + ' -> ' \
               + str(self.out_features) + ')'
               
class SimpleGCDEC(nn.Module):
    """Basic model used in SpaGCN training.

    Parameters
    ----------
    nfeat : int
        input feature dimension
    
    nhid : int
        output feature dimension

    alpha : float optional
        alpha, by default as 0.2
        
    """
    def __init__(self, nfeat, nhid, alpha=0.2):
        super(SimpleGCDEC, self).__init__()
        self.gc = GraphConvolution(nfeat, nhid)
        self.nhid=nhid
        #self.mu determined by the init method
        self.alpha=alpha

    def forward(self, x, adj):
        """forward function.

        Parameters
        ----------
        x :  
            node features.
        adj : 
            adjacent matrix.
       

        Returns
        -------
        x : 
            the output of graph convolution layer.
        q : 
            the probability of assigning cell i to cluster j.

        """
        x=self.gc(x, adj)
        q = 1.0 / ((1.0 + torch.sum((x.unsqueeze(1) - self.mu)**2, dim=2) / self.alpha) + 1e-8)
        q = q**(self.alpha+1.0)/2.0
        q = q / torch.sum(q, dim=1, keepdim=True)
        return x, q

    def loss_function(self, p, q):
        """objective function as a Kullback–Leibler (KL) divergence loss.

        Parameters
        ----------
        p :  
            target distribution.
        q : 
            the probability of assigning cell i to cluster j.
       

        Returns
        -------
        loss : 
            Kullback–Leibler (KL) divergence loss.

        """
        def kld(target, pred):
            return torch.mean(torch.sum(target*torch.log(target/(pred+1e-6)), dim=1))
        loss = kld(p, q)
        return loss

    def target_distribution(self, q):
        """generate an auxiliary target distribution based on q the probability of assigning cell i to cluster j.

        Parameters
        ----------
        q : 
            the probability of assigning cell i to cluster j.
        
        Returns
        -------
        p : 
            target distribution.

        """
        p = q**2 / torch.sum(q, dim=0)
        p = p / torch.sum(p, dim=1, keepdim=True)
        return p

    def fit(self, X, adj,  lr=0.001, max_epochs=5000, update_interval=3, trajectory_interval=50, weight_decay=5e-4, opt="sgd", init="louvain", n_neighbors=10, res=0.4, n_clusters=10, init_spa=True, tol=1e-3):
        """fit function for model training.

        Parameters
        ----------
        X :  
            node features.
        adj : 
            adjacent matrix.
        lr : float optional
            learning rate.
        max_epochs : int optional
            max epochs.
        update_interval: int optional
            interval for update
        trajectory_interval: int optional
            trajectory interval
        weight_decay : float optional
            weight decay.
        opt : str optional
            optimization.
        init : str optional
            "louvain" or "kmeans".
        n_neighbors : int optional
            the number of neighbors used in louvain.
        res : float optional
            used for louvain .
        n_clusters : int optional
            the number of clusters usedd in kmeans.
        init_spa : bool optional
            initialize spatial.
        tol : float optional
            tolerant value for searching l.
        
        Returns
        -------
        None. 

        """
        self.trajectory=[]
        if opt=="sgd":
            optimizer = optim.SGD(self.parameters(), lr=lr, momentum=0.9)
        elif opt=="admin":
            optimizer = optim.Adam(self.parameters(),lr=lr, weight_decay=weight_decay)

        features= self.gc(torch.FloatTensor(X),torch.FloatTensor(adj))
        #----------------------------------------------------------------        
        if init=="kmeans":
            print("Initializing cluster centers with kmeans, n_clusters known")
            self.n_clusters=n_clusters
            kmeans = KMeans(self.n_clusters, n_init=20)
            if init_spa:
                #------Kmeans use exp and spatial
                y_pred = kmeans.fit_predict(features.detach().numpy())
            else:
                #------Kmeans only use exp info, no spatial
                y_pred = kmeans.fit_predict(X)  #Here we use X as numpy
        elif init=="louvain":
            print("Initializing cluster centers with louvain, resolution = ", res)
            if init_spa:
                adata=sc.AnnData(features.detach().numpy())
            else:
                adata=sc.AnnData(X)
            sc.pp.neighbors(adata, n_neighbors=n_neighbors)
            # sc.tl.louvain(adata,resolution=res)
            sc.tl.leiden(adata, resolution=res, key_added='louvain')
            

            y_pred=adata.obs['louvain'].astype(int).to_numpy()
            self.n_clusters=len(np.unique(y_pred))
        #----------------------------------------------------------------
        y_pred_last = y_pred
        self.mu = Parameter(torch.Tensor(self.n_clusters, self.nhid))
        X=torch.FloatTensor(X)
        adj=torch.FloatTensor(adj)
        self.trajectory.append(y_pred)
        features=pd.DataFrame(features.detach().numpy(),index=np.arange(0,features.shape[0]))
        Group=pd.Series(y_pred,index=np.arange(0,features.shape[0]),name="Group")
        Mergefeature=pd.concat([features,Group],axis=1)
        cluster_centers=np.asarray(Mergefeature.groupby("Group").mean())
        
        self.mu.data.copy_(torch.Tensor(cluster_centers))
        self.train()
        for epoch in range(max_epochs):
            if epoch%update_interval == 0:
                _, q = self.forward(X,adj)
                p = self.target_distribution(q).data
            if epoch%10==0:
                print("Epoch ", epoch) 
            optimizer.zero_grad()
            z,q = self(X, adj)
            loss = self.loss_function(p, q)
            loss.backward()
            optimizer.step()
            if epoch%trajectory_interval == 0:
                self.trajectory.append(torch.argmax(q, dim=1).data.cpu().numpy())

            #Check stop criterion
            y_pred = torch.argmax(q, dim=1).data.cpu().numpy()
            delta_label = np.sum(y_pred != y_pred_last).astype(np.float32) / X.shape[0]
            y_pred_last = y_pred
            if epoch>0 and (epoch-1)%update_interval == 0 and delta_label < tol:
                print('delta_label ', delta_label, '< tol ', tol)
                print("Reach tolerance threshold. Stopping training.")
                print("Total epoch:", epoch)
                break

    def fit_with_init(self, X, adj, init_y, lr=0.001, max_epochs=5000, update_interval=1, weight_decay=5e-4,opt="sgd"):
        """Initializing cluster centers with kmeans."""
        print("Initializing cluster centers with kmeans.")
        if opt=="sgd":
            optimizer = optim.SGD(self.parameters(), lr=lr, momentum=0.9)
        elif opt=="admin":
            optimizer = optim.Adam(self.parameters(),lr=lr, weight_decay=weight_decay)
        X=torch.FloatTensor(X)
        adj=torch.FloatTensor(adj)
        features, _ = self.forward(X,adj)
        features=pd.DataFrame(features.detach().numpy(),index=np.arange(0,features.shape[0]))
        Group=pd.Series(init_y,index=np.arange(0,features.shape[0]),name="Group")
        Mergefeature=pd.concat([features,Group],axis=1)
        cluster_centers=np.asarray(Mergefeature.groupby("Group").mean())
        self.mu.data.copy_(torch.Tensor(cluster_centers))
        self.train()
        for epoch in range(max_epochs):
            if epoch%update_interval == 0:
                _, q = self.forward(torch.FloatTensor(X),torch.FloatTensor(adj))
                p = self.target_distribution(q).data
            X=torch.FloatTensor(X)
            adj=torch.FloatTensor(adj)
            optimizer.zero_grad()
            z,q = self(X, adj)
            loss = self.loss_function(p, q)
            loss.backward()
            optimizer.step()

    def predict(self, X, adj):
        """transform to float tensor"""
        z, q = self(torch.FloatTensor(X), torch.FloatTensor(adj))
        return z, q

# Not used in SpaGCN
class GC_DEC(nn.Module):
    def __init__(self, nfeat, nhid1,nhid2, n_clusters=None, dropout=0.5,alpha=0.2):
        super(GC_DEC, self).__init__()

        self.gc1 = GraphConvolution(nfeat, nhid1)
        self.gc2 = GraphConvolution(nhid1, nhid2)
        self.dropout = dropout
        self.mu = Parameter(torch.Tensor(n_clusters, nhid2))
        self.n_clusters=n_clusters
        self.alpha=alpha

    def forward(self, x, adj):
        x=self.gc1(x, adj)
        x = F.relu(x)
        x = F.dropout(x, self.dropout, training=True)
        x = self.gc2(x, adj)
        q = 1.0 / ((1.0 + torch.sum((x.unsqueeze(1) - self.mu)**2, dim=2) / self.alpha) + 1e-6)
        q = q**(self.alpha+1.0)/2.0
        q = q / torch.sum(q, dim=1, keepdim=True)
        return x, q

    def loss_function(self, p, q):
        def kld(target, pred):
            return torch.mean(torch.sum(target*torch.log(target/(pred+1e-6)), dim=1))
        loss = kld(p, q)
        return loss

    def target_distribution(self, q):
        #weight = q ** 2 / q.sum(0)
        #return torch.transpose((torch.transpose(weight,0,1) / weight.sum(1)),0,1)e
        p = q**2 / torch.sum(q, dim=0)
        p = p / torch.sum(p, dim=1, keepdim=True)
        return p

    def fit(self, X,adj, lr=0.001, max_epochs=10, update_interval=5, weight_decay=5e-4,opt="sgd",init="louvain",n_neighbors=10,res=0.4):
        self.trajectory=[]
        print("Initializing cluster centers with kmeans.")
        if opt=="sgd":
            optimizer = optim.SGD(self.parameters(), lr=lr, momentum=0.9)
        elif opt=="admin":
            optimizer = optim.Adam(self.parameters(),lr=lr, weight_decay=weight_decay)

        features, _ = self.forward(torch.FloatTensor(X),torch.FloatTensor(adj))
        #----------------------------------------------------------------
        if init=="kmeans":
            #Kmeans only use exp info, no spatial
            #kmeans = KMeans(self.n_clusters, n_init=20)
            #y_pred = kmeans.fit_predict(X)  #Here we use X as numpy
            #Kmeans use exp and spatial
            kmeans = KMeans(self.n_clusters, n_init=20)
            y_pred = kmeans.fit_predict(features.detach().numpy())
        elif init=="louvain":
            adata=sc.AnnData(features.detach().numpy())
            sc.pp.neighbors(adata, n_neighbors=n_neighbors)
            sc.tl.louvain(adata,resolution=res)
            y_pred=adata.obs['louvain'].astype(int).to_numpy()
        #----------------------------------------------------------------
        X=torch.FloatTensor(X)
        adj=torch.FloatTensor(adj)
        self.trajectory.append(y_pred)
        features=pd.DataFrame(features.detach().numpy(),index=np.arange(0,features.shape[0]))
        Group=pd.Series(y_pred,index=np.arange(0,features.shape[0]),name="Group")
        Mergefeature=pd.concat([features,Group],axis=1)
        cluster_centers=np.asarray(Mergefeature.groupby("Group").mean())
        
        self.mu.data.copy_(torch.Tensor(cluster_centers))
        self.train()
        for epoch in range(max_epochs):
            if epoch%update_interval == 0:
                _, q = self.forward(X,adj)
                p = self.target_distribution(q).data
            if epoch%100==0:
                print("Epoch ", epoch) 
            optimizer.zero_grad()
            z,q = self(X, adj)
            loss = self.loss_function(p, q)
            loss.backward()
            optimizer.step()
            self.trajectory.append(torch.argmax(q, dim=1).data.cpu().numpy())

    def fit_with_init(self, X,adj, init_y, lr=0.001, max_epochs=10, update_interval=1, weight_decay=5e-4,opt="sgd"):
        print("Initializing cluster centers with kmeans.")
        if opt=="sgd":
            optimizer = optim.SGD(self.parameters(), lr=lr, momentum=0.9)
        elif opt=="admin":
            optimizer = optim.Adam(self.parameters(),lr=lr, weight_decay=weight_decay)
        X=torch.FloatTensor(X)
        adj=torch.FloatTensor(adj)
        features, _ = self.forward(X,adj)
        features=pd.DataFrame(features.detach().numpy(),index=np.arange(0,features.shape[0]))
        Group=pd.Series(init_y,index=np.arange(0,features.shape[0]),name="Group")
        Mergefeature=pd.concat([features,Group],axis=1)
        cluster_centers=np.asarray(Mergefeature.groupby("Group").mean())
        self.mu.data.copy_(torch.Tensor(cluster_centers))
        self.train()
        for epoch in range(max_epochs):
            if epoch%update_interval == 0:
                _, q = self.forward(torch.FloatTensor(X),torch.FloatTensor(adj))
                p = self.target_distribution(q).data
            X=torch.FloatTensor(X)
            adj=torch.FloatTensor(adj)
            optimizer.zero_grad()
            z,q = self(X, adj)
            loss = self.loss_function(p, q)
            loss.backward()
            optimizer.step()

    def predict(self, X, adj):
        z,q = self(torch.FloatTensor(X),torch.FloatTensor(adj))
        return z, q

class SpaGCN(object):
    """SpaGCN class.

    Parameters
    ----------
    l : float
        the parameter to control percentage p
        
    """
    def __init__(self, l=None ):
        super(SpaGCN, self).__init__()
        self.l = l or None
        self.res = None

    def search_l(self, p, adj, start=0.01, end=1000, tol=0.01, max_run=100):
        """Search best l.

        Parameters
        ----------
        p : float 
            percentage.
        adj : 
            adjacent matrix.
        start : float optional
            starting value for searching l.
        end : float optional
            ending value for searching l.
        tol : float optional
            tolerant value for searching l.
        max_run : int optional
            Max runs.

        Returns
        -------
        l : float
            best l, the parameter to control percentage p.

        """
        l = utils.search_l(p, adj, start, end, tol, max_run)
        return l
    
    def set_l(self, l):
        """set l.

        Parameters
        ----------
        l : float 
            the parameter to control percentage p.
       
        Returns
        -------
        None. 
        
        """
        self.l = l

    def search_set_res(self, adata, adj, l, target_num, start=0.4, step=0.1, tol=5e-3, lr=0.05, max_epochs=10, r_seed=100, t_seed=100, n_seed=100, max_run=10):
        """search res. res: Resolution in the initial Louvain's Clustering methods.
        """
        random.seed(r_seed)
        torch.manual_seed(t_seed)
        np.random.seed(n_seed)
        res=start
        print("Start at res = ", res, "step = ", step)
        clf=SpaGCN()
        clf.set_l(l)
        clf.fit(adata,adj,init_spa=True,init="louvain",res=res, tol=tol, lr=lr, max_epochs=max_epochs)
        y_pred, _=clf.predict()
        old_num=len(set(y_pred))
        print("Res = ", res, "Num of clusters = ", old_num)
        run=0
        while old_num!=target_num:
            random.seed(r_seed)
            torch.manual_seed(t_seed)
            np.random.seed(n_seed)
            old_sign=1 if (old_num<target_num) else -1
            clf=SpaGCN()
            clf.set_l(l)
            clf.fit(adata,adj,init_spa=True,init="louvain",res=res+step*old_sign, tol=tol, lr=lr, max_epochs=max_epochs)
            y_pred, _=clf.predict()
            new_num=len(set(y_pred))
            print("Res = ", res+step*old_sign, "Num of clusters = ", new_num)
            if new_num==target_num:
                res=res+step*old_sign
                print("recommended res = ", str(res))
                return res
            new_sign=1 if (new_num<target_num) else -1
            if new_sign==old_sign:
                res=res+step*old_sign
                print("Res changed to", res)
                old_num=new_num
            else:
                step=step/2
                print("Step changed to", step)
            if run >max_run:
                print("Exact resolution not found")
                print("Recommended res = ", str(res))
                return res
            run+=1
        print("recommended res = ", str(res))
        self.res = res
        return res

    def fit(self,adata,adj, 
            num_pcs=50, 
            lr=0.005,
            max_epochs=2000,
            weight_decay=0,
            opt="admin",
            init_spa=True,
            init="louvain", #louvain or kmeans
            n_neighbors=10, #for louvain
            n_clusters=None, #for kmeans
            res=0.4, #for louvain
            tol=1e-3):
        """fit function for model training.

        Parameters
        ----------
        adata :  
            input data.
        adj : 
            adjacent matrix.
        num_pcs : int 
            the number of component used in PCA.
        lr : float
            learning rate.
        max_epochs : int
            max epochs.
        weight_decay : float
            weight decay.
        opt : str
            optimization.
        init_spa : bool
            initialize spatial.
        init : str
            "louvain" or "kmeans".
        n_neighbors : int
            the number of neighbors used in louvain.
        n_clusters : int
            the number of clusters usedd in kmeans.
        res : float
            used for louvain .
        tol : float
            tolerant value for searching l.
        
        Returns
        -------
        None. 

        """
        self.num_pcs=num_pcs
        self.res=res
        self.lr=lr
        self.max_epochs=max_epochs
        self.weight_decay=weight_decay
        self.opt=opt
        self.init_spa=init_spa
        self.init=init
        self.n_neighbors=n_neighbors
        self.n_clusters=n_clusters
        self.res=res
        self.tol=tol
        assert adata.shape[0]==adj.shape[0]==adj.shape[1]
        pca = PCA(n_components=self.num_pcs)
        if issparse(adata.X):
            pca.fit(adata.X.A)
            embed=pca.transform(adata.X.A)
        else:
            pca.fit(adata.X)
            embed=pca.transform(adata.X)
        ###------------------------------------------###
        if self.l is None:
            raise ValueError('l should be set before fitting the model!')
        adj_exp=np.exp(-1*(adj**2)/(2*(self.l**2)))
        #----------Train model----------
        self.model=SimpleGCDEC(embed.shape[1],embed.shape[1])
        self.model.fit(embed,adj_exp,lr=self.lr,max_epochs=self.max_epochs,weight_decay=self.weight_decay,opt=self.opt,init_spa=self.init_spa,init=self.init,n_neighbors=self.n_neighbors,n_clusters=self.n_clusters,res=self.res, tol=self.tol)
        self.embed=embed
        self.adj_exp=adj_exp

    def predict(self):
        """prediction function.

        Parameters
        ----------
     
        Returns
        -------
        y_pred : numpy
            predicted label.
        prob : numpy
            predicted probability.
        
        """
        z,q=self.model.predict(self.embed,self.adj_exp)
        y_pred = torch.argmax(q, dim=1).data.cpu().numpy()
        self.y_pred = y_pred
        # Max probability plot
        prob=q.detach().numpy()
        return y_pred, prob
    
    def score(self, y_true):
        """score function to get score of prediction.

        Parameters
        ----------
        y_true :  
            ground truth label.
            
        Returns
        -------
        score : float
            metric eval score.
        
        """
        from sklearn.metrics.cluster import adjusted_rand_score
        score = adjusted_rand_score(y_true, self.y_pred)
        print("ARI {}".format(adjusted_rand_score(y_true, self.y_pred)))
        return score



from dance.utils import set_seed
from dance.datasets.multimodality import JointEmbeddingNIPSDataset
from dance.modules.multi_modality.joint_embedding.scmogcnv2 import ScMoGCNWrapper
import anndata as ad
import argparse
import random
import torch
import scanpy as sc
import numpy as np
import dgl
from dance.transforms.graph_construct import construct_basic_feature_graph, gen_batch_features
import dance.utils.metrics as metrics

def scmogcn_test(adata):
    sc._settings.ScanpyConfig.n_jobs = 4
    adata_sol = dataset.test_sol
    adata.obs['batch'] = adata_sol.obs['batch'][adata.obs_names]
    adata.obs['cell_type'] = adata_sol.obs['cell_type'][adata.obs_names]
    print(adata.shape,adata_sol.shape)
    adata_bc = adata.obs_names
    adata_sol_bc = adata_sol.obs_names
    select = [item in adata_bc for item in adata_sol_bc]
    adata_sol = adata_sol[select, :]
    print(adata.shape, adata_sol.shape)

    adata.obsm['X_emb'] = adata.X
    nmi = metrics.get_nmi(adata)
    cell_type_asw = metrics.get_cell_type_ASW(adata)
    cc_con = metrics.get_cell_cycle_conservation(adata, adata_sol)
    traj_con = metrics.get_traj_conservation(adata, adata_sol)
    batch_asw = metrics.get_batch_ASW(adata)
    graph_score = metrics.get_graph_connectivity(adata)

    print('nmi %.4f, celltype_asw %.4f, cc_con %.4f, traj_con %.4f, batch_asw %.4f, graph_score %.4f\n' % (
    nmi, cell_type_asw, cc_con, traj_con, batch_asw, graph_score))

    print('average metric: %.4f' % np.mean(
        [round(i, 4) for i in [nmi, cell_type_asw, cc_con, traj_con, batch_asw, graph_score]]))

if __name__ == '__main__':
    rndseed = random.randint(0, 2147483647)
    parser = argparse.ArgumentParser()
    parser.add_argument('-t', '--subtask', default = 'openproblems_bmmc_cite_phase2', choices = ['openproblems_bmmc_cite_phase2', 'openproblems_bmmc_multiome_phase2'])
    parser.add_argument('-d', '--data_folder', default = './data/joint_embedding')
    parser.add_argument('-pre', '--pretrained_folder', default = './data/joint_embedding/pretrained')
    parser.add_argument('-csv', '--csv_path', default = 'decoupled_lsi.csv')
    parser.add_argument('-seed', '--rnd_seed', default = rndseed, type=int)
    parser.add_argument('-cpu', '--cpus', default = 1, type = int)
    parser.add_argument('-bs', '--batch_size', default = 512, type=int)
    parser.add_argument('-prefix', '--prefix', default='dance_openproblems_bmmc_atac2rna_test')

    parser.add_argument('-pww', '--pathway_weight', default='pearson', choices=['cos', 'one', 'pearson'])
    parser.add_argument('-pwth', '--pathway_threshold', type=float, default=-1.0)
    parser.add_argument('-l', '--log_folder', default='../../../single_cell/logs')
    parser.add_argument('-m', '--model_folder', default='../../../single_cell/models')
    parser.add_argument('-r', '--result_folder', default='../../../single_cell/results')
    parser.add_argument('-e', '--epoch', type=int, default=1000)
    parser.add_argument('-nbf', '--no_batch_features', action='store_true')
    parser.add_argument('-npw', '--no_pathway', action='store_true')
    parser.add_argument('-opw', '--only_pathway', action='store_true')
    parser.add_argument('-res', '--residual', default='res_cat', choices=['none', 'res_add', 'res_cat'])
    parser.add_argument('-inres', '--initial_residual', action='store_true')
    parser.add_argument('-pwagg', '--pathway_aggregation', default='alpha',
                        choices=['sum', 'attention', 'two_gate', 'one_gate', 'alpha', 'cat'])
    parser.add_argument('-pwalpha', '--pathway_alpha', type=float, default=0.5)
    parser.add_argument('-nrc', '--no_readout_concatenate', action='store_true')
    parser.add_argument('-nm', '--normalization', default='group', choices=['batch', 'layer', 'group', 'none'])
    parser.add_argument('-ac', '--activation', default='gelu', choices=['leaky_relu', 'relu', 'prelu', 'gelu'])
    parser.add_argument('-em', '--embedding_layers', default=1, type=int, choices=[1, 2, 3])
    parser.add_argument('-ro', '--readout_layers', default=2, type=int, choices=[1, 2, 3])
    parser.add_argument('-conv', '--conv_layers', default=4, type=int, choices=[1, 2, 3, 4, 5, 6])
    parser.add_argument('-agg', '--agg_function', default='mean', choices=['gcn', 'mean'])
    parser.add_argument('-device', '--device', default='cuda')
    parser.add_argument('-sb', '--save_best', action='store_true')
    parser.add_argument('-sf', '--save_final', action='store_true')
    parser.add_argument('-lr', '--learning_rate', type=float, default=1e-2)
    parser.add_argument('-lrd', '--lr_decay', type=float, default=0.99)
    parser.add_argument('-wd', '--weight_decay', type=float, default=1e-4)
    parser.add_argument('-hid', '--hidden_size', type=int, default=48)
    parser.add_argument('-edd', '--edge_dropout', type=float, default=0)
    parser.add_argument('-mdd', '--model_dropout', type=float, default=0.4)
    parser.add_argument('-es', '--early_stopping', type=int, default=20)
    parser.add_argument('-or', '--output_relu', default='none', choices=['relu', 'leaky_relu', 'none'])
    parser.add_argument('-i', '--inductive', default='trans', choices=['normal', 'opt', 'trans'])
    parser.add_argument('-sa', '--subpath_activation', action='store_true')
    parser.add_argument('-ci', '--cell_init', default='none', choices=['none', 'pca'])
    parser.add_argument('-bas', '--batch_seperation', action='store_true')
    parser.add_argument('-pwpath', '--pathway_path', default='../../../single_cell/task1/h.all.v7.4')
    parser.add_argument('-ws', '--weighted_sum', action='store_true')

    args = parser.parse_args()
    args.rnd_seed = 1507653565

    device = args.device
    torch.set_num_threads(args.cpus)
    rndseed = args.rnd_seed
    set_seed(rndseed)

    args.no_pathway = True
    args.no_batch_features = True#False
    args.weighted_sum = True
    args.hidden_size = 100


    dataset = JointEmbeddingNIPSDataset(args.subtask, data_dir=args.data_folder).load_data()\
        .load_metadata().load_sol().preprocess('aux', args.pretrained_folder)
    X = torch.cat([dataset.tensor_features(0), dataset.tensor_features(1)], 1)
    Y_train = dataset.preprocessed_data['Y_train']
    g = construct_basic_feature_graph(X, device=device)
    args.FEATURE_SIZE = X.shape[1]
    X = X.to(device)
    args.feat1 = dataset.modalities[0].shape[1]
    args.feat2 = dataset.modalities[1].shape[1]

    if not args.no_batch_features:
        batch_features = gen_batch_features([dataset.modalities[0]]).to(device)
        args.BATCH_NUM = batch_features.shape[1]
        g.nodes['cell'].data['bf'] = batch_features

    model = ScMoGCNWrapper(args)
    model.fit(g, X, Y_train, args.epoch)
    # model.load(f'models/model_joint_embedding_{rndseed}.pth')

    with torch.no_grad():
        embeds = model.predict(g).cpu().numpy()
        print(embeds)

    mod1_obs = dataset.modalities[0].obs
    mod1_uns = dataset.modalities[0].uns
    adata = ad.AnnData(
        X=embeds,
        obs=mod1_obs,
        uns={
            'dataset_id': mod1_uns['dataset_id'],
            'method_id': 'scmogcn',
        },
    )
    adata.write_h5ad(f'./joint_embedding_{rndseed}.h5ad', compression="gzip")

    # old Neurips metrics
    # scmogcn_test(adata)
    # scmvae test
    metrics.labeled_clustering_evaluate(adata, dataset)


import sys
import os.path as o
#use if running from dance/examples/spatial/cell_type_deconvo
root_path=o.abspath(o.join(o.dirname(sys.modules[__name__].__file__), "../../.."))
sys.path.append(root_path)

import pandas as pd
import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F
from torch import optim

from dance.datasets.spatial import CellTypeDeconvoDataset
from dance.modules.spatial.cell_type_deconvo.spatialdecon import SpatialDecon
device = torch.device("cuda" if torch.cuda.is_available() else "cpu")

# get data 
dataset = CellTypeDeconvoDataset("toy2", data_dir="data/spatial")

X=np.array(dataset.data['ref_sc_count'])
Y=np.array(dataset.data['mix_count'])

#set cell-types 
cell_types =np.array([np.repeat(i+1, X.shape[1]//3) for i in range(3)]).flatten()

def cell_topic_profile(X, groups,axis=0, method='median'):
    ids=np.unique(groups)
    if method == "median":
        X_profile = np.array([np.median(X[:, groups==ids[i]], axis=1) for i in range(len(ids))]).T
    else:
        X_profile = np.array([np.mean(X[:, groups==ids[i]], axis=1) for i in range(len(ids))]).T
    return X_profile
X_profile = cell_topic_profile(X, cell_types, 0, "mean" )


y=Y[:,:5]
print(y.shape)
x=X_profile
print(x.shape)

torch.manual_seed(17)
#run deconvolution with logNormReg (multiplicative log-normal errors)
deconLNR = SpatialDecon(in_dim=x.shape[1], out_dim=y.shape[1], device=device) #,sd=.13, mu=.4)
deconLNR.fit(x, y, max_iter=4000, lr=.0001)
proportion_preds = deconLNR.predict().clone().detach()
print('MSLE score:',deconLNR.score(x, y))
print('Predicted Proportions:', proportion_preds)


import warnings
warnings.filterwarnings("ignore")
import argparse
from time import time
import os
import random
import numpy as np
import scanpy as sc
import torch

from dance.transforms.preprocess import normalize_adata, generate_random_pair
from dance.modules.single_modality.clustering.scdcc import ScDCC
from dance.datasets.singlemodality import ClusteringDataset

# for repeatability
random.seed(42)
np.random.seed(42)
torch.manual_seed(42)
torch.cuda.manual_seed(42)

if __name__ == "__main__":

    # setting the hyper parameters
    parser = argparse.ArgumentParser(description='train',
                                     formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument('--label_cells', default=0.1, type=float)
    parser.add_argument('--label_cells_files', default='label_mouse_ES_cell.txt')
    parser.add_argument('--n_pairwise', default=0, type=int)
    parser.add_argument('--n_pairwise_error', default=0, type=float)
    parser.add_argument('--batch_size', default=256, type=int)
    parser.add_argument('--data_dir', default='./data')
    parser.add_argument('--data_file', default='mouse_ES_cell',
                        type=str)  # choice=['10X_PBMC', 'mouse_bladder_cell', 'mouse_ES_cell', 'worm_neuron_cell']
    parser.add_argument('--maxiter', default=500, type=int)
    parser.add_argument('--pretrain_epochs', default=50, type=int)
    parser.add_argument('--lr', default=0.1, type=float)
    parser.add_argument('--sigma', default=2.5, type=float,
                        help='coefficient of Gaussian noise')
    parser.add_argument('--gamma', default=1., type=float,
                        help='coefficient of clustering loss')
    parser.add_argument('--ml_weight', default=1., type=float,
                        help='coefficient of must-link loss')
    parser.add_argument('--cl_weight', default=1., type=float,
                        help='coefficient of cannot-link loss')
    parser.add_argument('--update_interval', default=1, type=int)
    parser.add_argument('--tol', default=0.0001, type=float)
    parser.add_argument('--ae_weights', default=None)
    parser.add_argument('--save_dir', default='results/scdcc/')
    parser.add_argument('--ae_weight_file', default='AE_weights.pth.tar')

    args = parser.parse_args()

    data = ClusteringDataset(args.data_dir, args.data_file).load_data()
    x = data.X
    y = data.Y
    n_clusters = len(np.unique(y))

    # preprocessing scRNA-seq read counts matrix
    adata = sc.AnnData(x)
    adata.obs['Group'] = y
    adata = adata.copy()
    adata.obs['DCA_split'] = 'train'
    adata.obs['DCA_split'] = adata.obs['DCA_split'].astype('category')

    adata = normalize_adata(adata,
                            size_factors=True,
                            normalize_input=True,
                            logtrans_input=True)

    input_size = adata.n_vars

    if not os.path.exists(args.label_cells_files):
        indx = np.arange(len(y))
        np.random.shuffle(indx)
        label_cell_indx = indx[0:int(np.ceil(args.label_cells * len(y)))]
    else:
        label_cell_indx = np.loadtxt(args.label_cells_files, dtype=np.int)

    x_sd = adata.X.std(0)
    x_sd_median = np.median(x_sd)

    if args.n_pairwise > 0:
        ml_ind1, ml_ind2, cl_ind1, cl_ind2, error_num = generate_random_pair(y, label_cell_indx, args.n_pairwise,
                                                                             args.n_pairwise_error)

        print("Must link paris: %d" % ml_ind1.shape[0])
        print("Cannot link paris: %d" % cl_ind1.shape[0])
        print("Number of error pairs: %d" % error_num)
    else:
        ml_ind1, ml_ind2, cl_ind1, cl_ind2 = np.array([]), np.array([]), np.array([]), np.array([])

    sigma = 2.75
    use_cuda = torch.cuda.is_available()
    if use_cuda:
        device = 'cuda'
    else:
        device = 'cpu'
    model = ScDCC(input_dim=adata.n_vars, z_dim=32, n_clusters=n_clusters,
                  encodeLayer=[256, 64], decodeLayer=[64, 256], sigma=args.sigma, gamma=args.gamma,
                  ml_weight=args.ml_weight, cl_weight=args.ml_weight).to(device)

    t0 = time()
    if args.ae_weights is None:
        model.pretrain_autoencoder(x=adata.X, X_raw=adata.raw.X, size_factor=adata.obs.size_factors,
                                   batch_size=args.batch_size, epochs=args.pretrain_epochs,
                                   ae_weights=args.ae_weight_file)
    else:
        if os.path.isfile(args.ae_weights):
            print("==> loading checkpoint '{}'".format(args.ae_weights))
            checkpoint = torch.load(args.ae_weights)
            model.load_state_dict(checkpoint['ae_state_dict'])
        else:
            print("==> no checkpoint found at '{}'".format(args.ae_weights))
            raise ValueError

    print('Pretraining time: %d seconds.' % int(time() - t0))

    if not os.path.exists(args.save_dir):
        os.makedirs(args.save_dir)

    model.fit(X=adata.X, X_raw=adata.raw.X, sf=adata.obs.size_factors, y=y, lr=args.lr, batch_size=args.batch_size,
              num_epochs=args.maxiter, ml_ind1=ml_ind1, ml_ind2=ml_ind2, cl_ind1=cl_ind1, cl_ind2=cl_ind2,
              update_interval=args.update_interval, tol=args.tol, save_dir=args.save_dir)
    print('Total time: %d seconds.' % int(time() - t0))

    y_pred = model.predict()
    #    print(f'Prediction: {y_pred}')

    acc, nmi, ari = model.score(y)
    print("ACC: %.4f, NMI: %.4f, ARI: %.4f" % (acc, nmi, ari))

    if not os.path.exists(args.label_cells_files):
        np.savetxt(args.label_cells_files, label_cell_indx, fmt="%i")

""" Reproduction information
10X PBMC:
python scdcc.py --data_file='10X_PBMC' --label_cells_files='label_10X_PBMC.txt' --gamma=1.5

Mouse ES:
python scdcc.py --data_file='mouse_ES_cell' --label_cells_files='label_mouse_ES_cell.txt' --gamma=1 --ml_weight=0.8 --cl_weight=0.8

Worm Neuron:
python scdcc.py --data_file='worm_neuron_cell' --label_cells_files='label_worm_neuron_cell.txt' --gamma=1

Mouse Bladder:
python scdcc.py --data_file='mouse_bladder_cell' --label_cells_files='label_mouse_bladder_cell.txt' --gamma=1.5
"""


import torch.nn as nn
import torch.nn.functional as F
import torch
import math
import torch.optim as optim
import numpy as np
import os
import dgl.nn as dglnn
import dgl
from torch.utils.data import DataLoader
from dance.utils.metrics import *
from dance.utils import SimpleIndexDataset

def propagation_layer_combination(X, idx, wt, from_logits=True):
    if from_logits:
        wt = torch.softmax(wt, -1)

    x = 0
    for i in range(wt.shape[0]):
        x += wt[i] * X[i][idx]

    return x

class ScMoGCN(nn.Module):
    def __init__(self, args):
        super().__init__()
        self.args = args
        self.opw = args.only_pathway
        self.npw = args.no_pathway
        self.nrc = args.no_readout_concatenate

        hid_feats = args.hidden_size
        out_feats = args.FEATURE_SIZE
        FEATURE_SIZE = args.FEATURE_SIZE

        if not args.no_batch_features:
            self.extra_encoder = nn.Linear(args.BATCH_NUM, hid_feats)

        if args.cell_init == 'none':
            self.embed_cell = nn.Embedding(2, hid_feats)
        else:
            self.embed_cell = nn.Linear(100, hid_feats)

        self.embed_feat = nn.Embedding(FEATURE_SIZE, hid_feats)

        self.input_linears = nn.ModuleList()
        self.input_acts = nn.ModuleList()
        self.input_norm = nn.ModuleList()
        for i in range((args.embedding_layers - 1) * 2):
            self.input_linears.append(nn.Linear(hid_feats, hid_feats))

        for i in range((args.embedding_layers - 1) * 2):
            self.input_acts.append(nn.GELU())

        for i in range((args.embedding_layers - 1) * 2):
            self.input_norm.append(nn.GroupNorm(4, hid_feats))

        if self.opw:
            self.edges = ['feature2cell', 'pathway']
        elif self.npw:
            self.edges = ['feature2cell', 'cell2feature']
        else:
            self.edges = ['feature2cell', 'cell2feature', 'pathway']

        self.conv_layers = nn.ModuleList()
        if args.residual == 'res_cat':
            self.conv_layers.append(dglnn.HeteroGraphConv(dict(zip(self.edges, [
                dglnn.SAGEConv(in_feats=hid_feats, out_feats=hid_feats, aggregator_type=args.agg_function, norm=None)
                for i in range(len(self.edges))])), aggregate='stack'))
            for i in range(args.conv_layers - 1):
                self.conv_layers.append(dglnn.HeteroGraphConv(dict(zip(self.edges, [
                    dglnn.SAGEConv(in_feats=hid_feats * 2, out_feats=hid_feats, aggregator_type=args.agg_function,
                                   norm=None) for i in range(len(self.edges))])), aggregate='stack'))

        else:
            for i in range(args.conv_layers):
                self.conv_layers.append(dglnn.HeteroGraphConv(dict(zip(self.edges, [
                    dglnn.SAGEConv(in_feats=hid_feats, out_feats=hid_feats, aggregator_type=args.agg_function,
                                   norm=None) for i in range(len(self.edges))])), aggregate='stack'))

        self.conv_acts = nn.ModuleList()
        self.conv_norm = nn.ModuleList()
        for i in range(args.conv_layers * 2):
            self.conv_acts.append(nn.GELU())

        for i in range(args.conv_layers * len(self.edges)):
            self.conv_norm.append(nn.GroupNorm(4, hid_feats))

        self.att_linears = nn.ModuleList()
        for i in range(args.conv_layers):
            self.att_linears.append(nn.Linear(hid_feats * 2, hid_feats))

        self.decoder = nn.ModuleList()
        self.decoder_acts = nn.ModuleList()

        if args.weighted_sum:
            print("Weighted_sum enabled. Argument '--no_readout_concatenate' won't take effect.")
            for i in range(args.readout_layers - 1):
                self.decoder.append(nn.Linear(hid_feats, hid_feats))
            self.decoder.append(nn.Linear(hid_feats, out_feats))
        elif self.nrc:
            for i in range(args.readout_layers - 1):
                self.decoder.append(nn.Linear(hid_feats, hid_feats))
            self.decoder.append(nn.Linear(hid_feats, out_feats))
        else:
            for i in range(args.readout_layers - 1):
                self.decoder.append(nn.Linear(hid_feats * args.conv_layers, hid_feats * args.conv_layers))
            self.decoder.append(nn.Linear(hid_feats * args.conv_layers, out_feats))


        for i in range(args.readout_layers - 1):
            self.decoder_acts.append(nn.GELU())

        self.wt = nn.Parameter(torch.zeros(args.conv_layers))
        if args.pathway_aggregation == 'alpha' and args.pathway_alpha < 0:
            self.aph = nn.Parameter(torch.zeros(2))

    def attention_agg(self, layer, h0, h):
        # h: h^{l-1}, dimension: (batch, hidden)
        # feats: result from two conv(cell conv and pathway conv), stacked together; dimension: (batch, 2, hidden)
        args = self.args
        if h.shape[1] == 1:
            return self.conv_norm[layer * len(self.edges) + 1](h.squeeze(1))
        elif args.pathway_aggregation == 'sum':
            return h[:, 0, :] + h[:, 1, :]
        else:
            h1 = h[:, 0, :]
            h2 = h[:, 1, :]

            if args.subpath_activation:
                h1 = F.leaky_relu(h1)
                h2 = F.leaky_relu(h2)

            h1 = self.conv_norm[layer * len(self.edges) + 1](h1)
            h2 = self.conv_norm[layer * len(self.edges) + 2](h2)

        if args.pathway_aggregation == 'attention':
            feats = torch.stack([h1, h2], 1)
            att = torch.transpose(F.softmax(torch.matmul(feats, self.att_linears[layer](h0).unsqueeze(-1)), 1), 1, 2)
            feats = torch.matmul(att, feats)
            return feats.squeeze(1)
        elif args.pathway_aggregation == 'one_gate':
            att = torch.sigmoid(self.att_linears[layer](torch.cat([h0, h1, h2], 1)))
            return att * h1 + (1 - att) * h2
        elif args.pathway_aggregation == 'two_gate':
            att1 = torch.sigmoid(self.att_linears[layer * 2](torch.cat([h0, h1], 1)))
            att2 = torch.sigmoid(self.att_linears[layer * 2 + 1](torch.cat([h0, h2], 1)))
            return att1 * h1 + att2 * h2
        elif args.pathway_aggregation == 'alpha':
            if args.pathway_alpha < 0:
                weight = torch.softmax(self.aph, -1)
                return weight[0] * h1 + weight[1] * h2
            else:
                return (1 - args.pathway_alpha) * h1 + args.pathway_alpha * h2
        elif args.pathway_aggregation == 'cat':
            return self.att_linears[layer](torch.cat([h1, h2], 1))

    def conv(self, graph, layer, h, hist):
        args = self.args
        h0 = hist[-1]
        h = self.conv_layers[layer](graph, h, mod_kwargs=dict(zip(self.edges, [{'edge_weight': F.dropout(
            graph.edges[self.edges[i]].data['weight'], p=args.edge_dropout, training=self.training)} for i in
                                                                               range(len(self.edges))])))

        if args.model_dropout > 0:
            h = {'feature': F.dropout(self.conv_acts[layer * 2](self.attention_agg(layer, h0['feature'], h['feature'])),
                                      p=args.model_dropout, training=self.training),
                 'cell': F.dropout(
                     self.conv_acts[layer * 2 + 1](self.conv_norm[layer * len(self.edges)](h['cell'].squeeze(1))),
                     p=args.model_dropout, training=self.training)}
        else:
            h = {'feature': self.conv_acts[layer * 2](self.attention_agg(layer, h0['feature'], h['feature'])),
                 'cell': self.conv_acts[layer * 2 + 1](self.conv_norm[layer * len(self.edges)](h['cell'].squeeze(1)))}

        return h

    def calculate_initial_embedding(self, graph):
        args = self.args

        input1 = F.leaky_relu(self.embed_feat(graph.srcdata['id']['feature']))
        input2 = F.leaky_relu(self.embed_cell(graph.srcdata['id']['cell']))

        if not args.no_batch_features:
            batch_features = graph.srcdata['bf']['cell']
            input2 += F.leaky_relu(F.dropout(self.extra_encoder(batch_features), p=0.2, training=self.training))[
                      :input2.shape[0]]

        hfeat = input1
        hcell = input2
        for i in range(args.embedding_layers - 1, (args.embedding_layers - 1) * 2):
            hfeat = self.input_linears[i](hfeat)
            hfeat = self.input_acts[i](hfeat)
            if args.normalization != 'none':
                hfeat = self.input_norm[i](hfeat)
            if args.model_dropout > 0:
                hfeat = F.dropout(hfeat, p=args.model_dropout, training=self.training)

        for i in range(args.embedding_layers - 1):
            hcell = self.input_linears[i](hcell)
            hcell = self.input_acts[i](hcell)
            if args.normalization != 'none':
                hcell = self.input_norm[i](hcell)
            if args.model_dropout > 0:
                hcell = F.dropout(hcell, p=args.model_dropout, training=self.training)

        return hfeat, hcell

    def propagate_with_sampling(self, blocks):
        args = self.args
        hfeat, hcell = self.calculate_initial_embedding(blocks[0])

        h = {'feature': hfeat, 'cell': hcell}

        for i in range(args.conv_layers):

            if i > 0:
                hfeat0, hcell0 = self.calculate_initial_embedding(blocks[i])

                h = {'feature': torch.cat([h['feature'], hfeat0], 1),
                     'cell': torch.cat([h['cell'], hcell0], 1)}

            hist = [h]
            h = self.conv(blocks[i], i, h, hist)

        hist = [h] * (args.conv_layers + 1)
        return hist  # , hist[-1]['feature']

    def propagate(self, graph):
        args = self.args
        hfeat, hcell = self.calculate_initial_embedding(graph)

        h = {'feature': hfeat, 'cell': hcell}
        hist = [h]

        for i in range(args.conv_layers):
            if i == 0 or args.residual == 'none':
                pass
            elif args.residual == 'res_add':
                if args.initial_residual:
                    h = {'feature': h['feature'] + hist[0]['feature'],
                         'cell': h['cell'] + hist[0]['cell']}

                else:
                    h = {'feature': h['feature'] + hist[-2]['feature'],
                         'cell': h['cell'] + hist[-2]['cell']}

            elif args.residual == 'res_cat':
                if args.initial_residual:
                    h = {'feature': torch.cat([h['feature'], hist[0]['feature']], 1),
                         'cell': torch.cat([h['cell'], hist[0]['cell']], 1)}
                else:
                    h = {'feature': torch.cat([h['feature'], hist[-2]['feature']], 1),
                         'cell': torch.cat([h['cell'], hist[-2]['cell']], 1)}

            h = self.conv(graph, i, h, hist)
            hist.append(h)

        return hist  # , hist[-1]['feature']

    def encode(self, graph, sampled=False):
        args = self.args
        if sampled:
            hist = self.propagate_with_sampling(graph)
        else:
            hist = self.propagate(graph)

        if args.weighted_sum:
            h = 0
            weight = torch.softmax(self.wt, -1)
            for i in range(args.conv_layers):
                h += weight[i] * hist[i + 1]['cell']
        elif not self.nrc:
            h = torch.cat([i['cell'] for i in hist[1:]], 1)
        else:
            h = hist[-1]['cell']

        return h#(h-torch.mean(h, 1, False).reshape(-1,1))/torch.std(h, 1, False).reshape(-1,1) #F.normalize(h)

    def forward(self, graph, sampled=False):
        return self.encode(graph, sampled)

    def decode(self, h):

        for i in range(self.args.readout_layers - 1):
            h = self.decoder[i](h)
            h = F.dropout(self.decoder_acts[i](h), p=self.args.model_dropout, training=self.training)
        h = self.decoder[-1](h)

        return h

class ScMoGCNWrapper(object):
    """ScMoGCN class.

    Parameters
    ----------
    args : argparse.Namespace
        A Namespace object that contains arguments of ScMoGCN. For details of parameters in parser args, please refer to link (parser help document).
    """

    def __init__(self, args):
        super(ScMoGCNWrapper, self).__init__()
        self.args = args
        self.model = ScMoGCN(args).to(args.device)

    def fit(self, g, y, train_labels=None, epochs=500):
        """fit function for training.

        Parameters
        ----------
        g : dgl.DGLGraph
            Constructed cell-feature graph.
        y : torch.Tensor
            Modality features, used as labels for reconstruction.
        train_labels : torch.Tensor
            Training supervision, by default to be None.
        epochs : int optional
            Maximum number of training epochs, by default to be 500.

        Returns
        -------
        None.
        """

        g = g.long()
        idx = np.random.permutation(train_labels[0].shape[0])
        train_idx = idx[:int(idx.shape[0] * 0.9)]
        val_idx = idx[int(idx.shape[0] * 0.9):]

        train_dataset = SimpleIndexDataset(train_idx)
        train_loader = DataLoader(dataset=train_dataset,
                                  batch_size=5000, #self.args.batch_size,
                                  shuffle=True,
                                  num_workers=1, )

        mse = nn.MSELoss()

        if train_labels is not None:
            c_decoder = nn.Linear(self.args.hidden_size, train_labels[0].max()+1).to(self.args.device)
            b_decoder = nn.Linear(self.args.hidden_size, train_labels[1].max()+1).to(self.args.device)
            cc_decoder = nn.Linear(self.args.hidden_size, train_labels[3].shape[1]).to(self.args.device)
            train_labels = [torch.from_numpy(i).to(self.args.device) for i in train_labels]
            ce = nn.CrossEntropyLoss()
            optimizer = torch.optim.AdamW([{'params': self.model.parameters()},
                                           {'params': c_decoder.parameters()},
                                           {'params': b_decoder.parameters()},
                                           {'params': cc_decoder.parameters()}], lr=self.args.learning_rate)
        else:
            optimizer = torch.optim.AdamW(self.model.parameters(), lr=self.args.learning_rate)

        vals = []
        feature_weight = g.in_degrees(etype='cell2feature').float()

        for epoch in range(epochs):
            self.model.train()
            print('epoch', epoch)

            # loss = mse(self.model.decode(self.model.encode(g))[train_idx], y[train_idx])
            # optimizer.zero_grad()
            # loss.backward()
            # optimizer.step()

            total_loss = 0
            for iter, batch_idx in enumerate(train_loader):
                if True:
                    feature_sampled = torch.multinomial(feature_weight, int(0.6*len(g.nodes('feature'))), replacement=False).to(self.args.device)
                    subgraph = dgl.node_subgraph(g, {'cell': batch_idx.to(self.args.device), 'feature': feature_sampled})
                    # subgraph = g
                    emb = self.model.encode(subgraph)
                    output = self.model.decode(emb)#[batch_idx]
                else:
                    emb = self.model.encode(subgraph)
                    output = self.model.decode(emb)[batch_idx]

                # loss = mse(output, y[batch_idx])
                l1 = mse(output[:, :self.args.feat1], y[batch_idx, :self.args.feat1])
                l2 = mse(output[:, -self.args.feat2:], y[batch_idx, -self.args.feat2:])
                loss = l1 * 0.5 + l2 * 0.5

                if train_labels is not None:
                    # l3 = ce(c_decoder(emb), train_labels[0][batch_idx].long())
                    # l4 = ce(b_decoder(emb), train_labels[1][batch_idx].long())
                    # l5 = mse(cc_decoder(emb), train_labels[3][batch_idx].float())
                    l3 = ce(emb[:, :train_labels[0].max()+1], train_labels[0][batch_idx].long())
                    l4 = ce(emb[:, train_labels[0].max()+1:train_labels[0].max()+train_labels[1].max()+2], train_labels[1][batch_idx].long())
                    l5 = mse(cc_decoder(emb), train_labels[3][batch_idx].float())
                    loss += l3 * 0.3 + l4 * 0.05 + l5 * 0.05
                    del l3, l4, l5
                total_loss += math.sqrt(loss.item())

                optimizer.zero_grad()
                loss.backward()
                optimizer.step()

                del subgraph, output, loss, emb, l1, l2
                torch.cuda.empty_cache()
            print(f'train loss: ', total_loss / len(train_loader))

            vals.append(self.score(g, y, val_idx))
            print('val loss', vals[-1])

            if min(vals) == vals[-1]:
                if not os.path.exists('models'):
                    os.mkdir('models')
                torch.save(self.model.state_dict(), f'models/model_joint_embedding_{self.args.rnd_seed}.pth')

            if epoch>self.args.early_stopping and min(vals) != min(vals[-self.args.early_stopping:]):
                print('Early stopped.')
                break

            if epoch > 150:
                for p in optimizer.param_groups:
                    p['lr'] *= self.args.lr_decay

        return self.model

    def to(self, device):
        """Performs device conversion.

        Parameters
        ----------
        device : str
            Target device.

        Returns
        -------
        self : ScMoGCNWrapper
            Converted model.
        """
        self.args.device = device
        self.model = self.model.to(device)
        return self

    def load(self, path, map_location = None):
        """Load model parameters from checkpoint file.

        Parameters
        ----------
        path : str
            Path to the checkpoint file.
        map_location : str optional
            Mapped device. This parameter will be passed to torch.load function if not none.

        Returns
        -------
        None.

        """

        if map_location is not None:
            self.model.load_state_dict(torch.load(path, map_location=map_location))
        else:
            self.model.load_state_dict(torch.load(path))

    def predict(self, g):
        """Predict function to get latent representation of data.

        Parameters
        ----------
        g : dgl.DGLGraph
            Constructed cell-feature graph.

        Returns
        -------
        prediction : torch.Tensor
            Joint embedding of input data.

        """

        self.model.eval()
        with torch.no_grad():
            h = self.model.encode(g)
            # prediction = (h - torch.mean(h, 1, False).reshape(-1, 1)) / torch.std(h, 1, False).reshape(-1, 1)
            # return prediction
            return h

    def score(self, g, y, idx):
        """Score function to get score of prediction.

        Parameters
        ----------
        g : dgl.DGLGraph
            Constructed cell-feature graph.
        y : torch.Tensor
            Modality features, used as labels for reconstruction.
        idx : Iterable[int]
            Index of testing samples for scoring.

        Returns
        -------
        loss : float
            Reconstruction loss.

        """

        self.model.eval()
        mse = nn.MSELoss()
        #
        with torch.no_grad():
            output = self.model.decode(self.model.encode(g))
            l1 = mse(output[idx, :self.args.feat1], y[idx, :self.args.feat1])
            l2 = mse(output[idx, -self.args.feat2:], y[idx, -self.args.feat2:])
            loss = l1 * 0.5 + l2 * 0.5
            loss = math.sqrt(loss.item())
            # loss = mse(self.model.decode(self.model.encode(g))[idx], y[idx])
            return loss
        # return loss1, loss2, loss3, loss4



"""
This is the implementation of scDSC.
Extended from https://github.com/DHUDBlab/scDSC

Reference
----------
Gan, Yanglan, et al. "Deep structural clustering for single-cell RNA-seq data jointly through autoencoder and graph neural network." Briefings in Bioinformatics 23.2 (2022): bbac018.
"""

import numpy as np
from tqdm import tqdm
import math
from sklearn import metrics
import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.nn.parameter import Parameter
from torch.optim import Adam
from torch.nn import Linear
from torch.optim.optimizer import Optimizer
from torch.utils.data import DataLoader

from dance.utils.loss import ZINBLoss
from dance.utils.metrics import cluster_acc
from dance.transforms.preprocess import load_graph


class SCDSCWrapper(object):
    """scDSC wrapper class.

    Parameters
    ----------
    args : argparse.Namespace
        a Namespace contains arguments of scDSC. For details of parameters in parser args, please refer to link (parser help document).

    """

    def __init__(self, args):
        super(SCDSCWrapper, self).__init__()
        self.args = args
        self.device = args.device
        self.model = SCDSC(args).to(self.device)
        self.model_pre = AE(n_enc_1=args.n_enc_1, n_enc_2=args.n_enc_2, n_enc_3=args.n_enc_3,
                            n_dec_1=args.n_dec_1, n_dec_2=args.n_dec_2, n_dec_3=args.n_dec_3,
                            n_input=args.n_input, n_z1=args.n_z1, n_z2=args.n_z2, n_z3=args.n_z3).to(self.device)

    def target_distribution(self, q):
        """Calculate auxiliary target distribution p with q.

        Parameters
        ----------
        q :
            soft label.

        Returns
        -------
        p :
            target distribution.

        """
        p = q ** 2 / q.sum(0)
        return (p.t() / p.sum(1)).t()

    def pretrain_ae(self, dataset, batch_size, n_epochs, fname, lr=1e-3):
        """Pretrain autoencoder.

        Parameters
        ----------
        dataset :
            input dataset.
        batch_size : int
            size of batch.
        n_epochs : int
            number of epochs.
        lr : float optional
            learning rate.
        fname : str
            path to save autoencoder weights.

        Returns
        -------
        None.

        """
        device = self.device
        train_loader = DataLoader(dataset, batch_size, shuffle=True)
        model = self.model_pre
        optimizer = Adam(model.parameters(), lr=lr)
        for epoch in range(n_epochs):
            for batch_idx, (x, _) in enumerate(train_loader):
                x = x.to(device)
                x_bar, _, _, _, _, _, _, _ = model(x)

                x_bar = x_bar.cpu()
                x = x.cpu()
                loss = F.mse_loss(x_bar, x)
                optimizer.zero_grad()
                loss.backward()
                optimizer.step()

            with torch.no_grad():
                x = torch.Tensor(dataset.x).to(device).float()
                x_bar, _, _, _, z3, _, _, _ = model(x)
                loss = F.mse_loss(x_bar, x)
                print('Pretrain epoch %3d, MSE loss: %.8f' % (epoch + 1, loss))

        torch.save(model.state_dict(), fname)

    def fit(self, dataset, X_raw, sf, graph_path, lr=1e-03, n_epochs=300, bcl=0.1, cl=0.01, rl=1, zl=0.1):
        """Train model.

        Parameters
        ----------
        dataset :
            input dataset.
        X_raw :
            raw input features.
        sf : list
            scale factor of ZINB loss.
        graph_path : str
            path of graph file.
        lr : float optional
            learning rate.
        n_epochs : int optional
            number of epochs.
        bcl : float optional
            parameter of binary crossentropy loss.
        cl : float optional
            parameter of Kullback–Leibler divergence loss.
        rl : float optional
            parameter of reconstruction loss.
        zl : float optional
            parameter of ZINB loss.

        Returns
        -------
        None.

        """
        device = self.device
        model = self.model
        optimizer = Adam(model.parameters(), lr=lr)
        # optimizer = RAdam(model.parameters(), lr=lr)

        adj = load_graph(graph_path, dataset.x)
        adj = adj.to(device)
        X_raw = torch.tensor(X_raw).to(device)
        sf = torch.tensor(sf).to(device)
        data = torch.Tensor(dataset.x).to(device)
        y = dataset.y

        aris = []
        P = {}
        Q = {}

        with torch.no_grad():
            _, _, _, _, z, _, _, _ = model.ae(data)

        for epoch in tqdm(range(n_epochs)):
            if epoch % 1 == 0:
                x_bar, tmp_q, pred, z, meanbatch, dispbatch, pibatch, zinb_loss = model(data, adj)
                tmp_q = tmp_q.data
                self.q = tmp_q
                p = self.target_distribution(tmp_q)

                # calculate ari score for model selection
                _, _, ari = self.score(y)
                aris.append(ari)
                print("Epoch %3d, ARI: %.4f, Best ARI: %.4f" % (epoch + 1, ari, max(aris)))

                p_ = {f'epoch{epoch}': p}
                q_ = {f'epoch{epoch}': tmp_q}
                P = {**P, **p_}
                Q = {**Q, **q_}

            x_bar, q, pred, z, meanbatch, dispbatch, pibatch, zinb_loss = model(data, adj)

            binary_crossentropy_loss = F.binary_cross_entropy(q, p)
            ce_loss = F.kl_div(pred.log(), p, reduction='batchmean')
            re_loss = F.mse_loss(x_bar, data)
            zinb_loss = zinb_loss(X_raw, meanbatch, dispbatch, pibatch, sf)
            loss = bcl * binary_crossentropy_loss + cl * ce_loss + rl * re_loss + zl * zinb_loss

            optimizer.zero_grad()
            loss.backward()
            optimizer.step()

        index = np.argmax(aris)
        self.q = Q[f'epoch{index}']

    def predict(self):
        """Get predictions from the trained model.

        Parameters
        ----------
        None.

        Returns
        -------
        y_pred : np.array
            prediction of given clustering method.

        """
        y_pred = torch.argmax(self.q, dim=1).data.cpu().numpy()
        return y_pred

    def score(self, y):
        """Evaluate the trained model.

        Parameters
        ----------
        y : list
            true labels.

        Returns
        -------
        acc : float
            accuracy.
        nmi : float
            normalized mutual information.
        ari : float
            adjusted Rand index.
        """
        y_pred = torch.argmax(self.q, dim=1).data.cpu().numpy()
        acc = np.round(cluster_acc(y, y_pred), 5)
        nmi = np.round(metrics.normalized_mutual_info_score(y, y_pred), 5)
        ari = np.round(metrics.adjusted_rand_score(y, y_pred), 5)
        return acc, nmi, ari


class SCDSC(nn.Module):
    """scDSC class.

    Parameters
    ----------
    args : argparse.Namespace
        a Namespace contains arguments of GCNAE. For details of parameters in parser args, please refer to link (parser help document).
    device : str
        computing device.
    sigma : float
        balance parameter.
    pretrain_path : str
        path of saved autoencoder weights.
    n_enc_1 : int
        output dimension of encoder layer 1.
    n_enc_2 : int
        output dimension of encoder layer 2.
    n_enc_3 : int
        output dimension of encoder layer 3.
    n_dec_1 : int
        output dimension of decoder layer 1.
    n_dec_2 : int
        output dimension of decoder layer 2.
    n_dec_3 : int
        output dimension of decoder layer 3.
    n_z1 : int
        output dimension of hidden layer 1.
    n_z2 : int
        output dimension of hidden layer 2.
    n_z3 : int
        output dimension of hidden layer 3.
    n_clusters : int
        number of clusters.
    n_input : int
        input feature dimension.
    v : float
        parameter of soft assignment.

    """

    def __init__(self, args):
        super(SCDSC, self).__init__()
        device = args.device
        self.sigma = args.sigma
        self.pretrain_path = args.pretrain_path
        self.ae = AE(
            n_enc_1=args.n_enc_1,
            n_enc_2=args.n_enc_2,
            n_enc_3=args.n_enc_3,

            n_dec_1=args.n_dec_1,
            n_dec_2=args.n_dec_2,
            n_dec_3=args.n_dec_3,

            n_input=args.n_input,
            n_z1=args.n_z1,
            n_z2=args.n_z2,
            n_z3=args.n_z3,
        )
        self.gnn_1 = GNNLayer(args.n_input, args.n_enc_1)
        self.gnn_2 = GNNLayer(args.n_enc_1, args.n_enc_2)
        self.gnn_3 = GNNLayer(args.n_enc_2, args.n_enc_3)
        self.gnn_4 = GNNLayer(args.n_enc_3, args.n_z1)
        self.gnn_5 = GNNLayer(args.n_z1, args.n_z2)
        self.gnn_6 = GNNLayer(args.n_z2, args.n_z3)
        self.gnn_7 = GNNLayer(args.n_z3, args.n_clusters)

        # cluster layer
        self.cluster_layer = Parameter(torch.Tensor(args.n_clusters, args.n_z3))
        torch.nn.init.xavier_normal_(self.cluster_layer.data)
        self._dec_mean = nn.Sequential(nn.Linear(args.n_dec_3, args.n_input), MeanAct())
        self._dec_disp = nn.Sequential(nn.Linear(args.n_dec_3, args.n_input), DispAct())
        self._dec_pi = nn.Sequential(nn.Linear(args.n_dec_3, args.n_input), nn.Sigmoid())
        # degree
        self.v = args.v
        self.zinb_loss = ZINBLoss().to(device)

    def forward(self, x, adj):
        """Forward propagation.

        Parameters
        ----------
        x :
            input features.
        adj :
            adjacency matrix

        Returns
        -------
        x_bar:
            reconstructed features.
        q :
            soft label.
        predict:
            prediction given by softmax assignment of embedding of GCN module
        z3 :
            embedding of autoencoder.
        _mean :
            data mean from ZINB.
        _disp :
            data dispersion from ZINB.
        _pi :
            data dropout probability from ZINB.
        zinb_loss:
            ZINB loss class.

        """
        # DNN Module
        self.ae.load_state_dict(torch.load(self.pretrain_path, map_location='cpu'))
        x_bar, tra1, tra2, tra3, z3, z2, z1, dec_h3 = self.ae(x)

        sigma = self.sigma
        # GCN Module
        h = self.gnn_1(x, adj)
        h = self.gnn_2((1 - sigma) * h + sigma * tra1, adj)
        h = self.gnn_3((1 - sigma) * h + sigma * tra2, adj)
        h = self.gnn_4((1 - sigma) * h + sigma * tra3, adj)
        h = self.gnn_5((1 - sigma) * h + sigma * z1, adj)
        h = self.gnn_6((1 - sigma) * h + sigma * z2, adj)
        h = self.gnn_7((1 - sigma) * h + sigma * z3, adj, active=False)

        predict = F.softmax(h, dim=1)

        _mean = self._dec_mean(dec_h3)
        _disp = self._dec_disp(dec_h3)
        _pi = self._dec_pi(dec_h3)
        zinb_loss = self.zinb_loss

        q = 1.0 / (1.0 + torch.sum(torch.pow(z3.unsqueeze(1) - self.cluster_layer, 2), 2) / self.v)
        q = q.pow((self.v + 1.0) / 2.0)
        q = (q.t() / torch.sum(q, 1)).t()

        return x_bar, q, predict, z3, _mean, _disp, _pi, zinb_loss


class GNNLayer(nn.Module):
    """GNN layer class. Construct a GNN layer with corresponding dimensions.

    Parameters
    ----------
    in_features : int
        input dimension of GNN layer.
    out_features : int
        output dimension of GNN layer.

    """

    def __init__(self, in_features, out_features):
        super(GNNLayer, self).__init__()
        self.in_features = in_features
        self.out_features = out_features
        self.weight = Parameter(torch.FloatTensor(in_features, out_features))
        torch.nn.init.xavier_uniform_(self.weight)
        # When passing through the network layer, the input and output variances are the same
        # including forward propagation and backward propagation

    def forward(self, features, adj, active=True):
        support = torch.mm(features, self.weight)
        output = torch.spmm(adj, support)
        if active:
            output = F.relu(output)
        return output


class AE(nn.Module):
    """Autoencoder class.

    Parameters
    ----------
    n_enc_1 : int
        output dimension of encoder layer 1.
    n_enc_2 : int
        output dimension of encoder layer 2.
    n_enc_3 : int
        output dimension of encoder layer 3.
    n_dec_1 : int
        output dimension of decoder layer 1.
    n_dec_2 : int
        output dimension of decoder layer 2.
    n_dec_3 : int
        output dimension of decoder layer 3.
    n_input : int
        input feature dimension.
    n_z1 : int
        output dimension of hidden layer 1.
    n_z2 : int
        output dimension of hidden layer 2.
    n_z3 : int
        output dimension of hidden layer 3.

    """

    def __init__(self, n_enc_1, n_enc_2, n_enc_3, n_dec_1, n_dec_2, n_dec_3,
                 n_input, n_z1, n_z2, n_z3):
        super(AE, self).__init__()

        self.enc_1 = Linear(n_input, n_enc_1)
        self.BN1 = nn.BatchNorm1d(n_enc_1)
        self.enc_2 = Linear(n_enc_1, n_enc_2)
        self.BN2 = nn.BatchNorm1d(n_enc_2)
        self.enc_3 = Linear(n_enc_2, n_enc_3)
        self.BN3 = nn.BatchNorm1d(n_enc_3)

        self.z1_layer = Linear(n_enc_3, n_z1)
        self.BN4 = nn.BatchNorm1d(n_z1)
        self.z2_layer = Linear(n_z1, n_z2)
        self.BN5 = nn.BatchNorm1d(n_z2)
        self.z3_layer = Linear(n_z2, n_z3)
        self.BN6 = nn.BatchNorm1d(n_z3)

        self.dec_1 = Linear(n_z3, n_dec_1)
        self.BN7 = nn.BatchNorm1d(n_dec_1)
        self.dec_2 = Linear(n_dec_1, n_dec_2)
        self.BN8 = nn.BatchNorm1d(n_dec_2)
        self.dec_3 = Linear(n_dec_2, n_dec_3)
        self.BN9 = nn.BatchNorm1d(n_dec_3)
        self.x_bar_layer = Linear(n_dec_3, n_input)

    def forward(self, x):
        """Forward propagation.

        Parameters
        ----------
        x :
            input features.

        Returns
        -------
        x_bar:
            reconstructed features.
        enc_h1:
            output of encoder layer 1.
        enc_h2:
            output of encoder layer 2.
        enc_h3:
            output of encoder layer 3.
        z3 :
            output of hidden layer 3.
        z2 :
            output of hidden layer 2.
        z1 :
            output of hidden layer 1.
        dec_h3 :
            output of decoder layer 3.

        """
        enc_h1 = F.relu(self.BN1(self.enc_1(x)))
        enc_h2 = F.relu(self.BN2(self.enc_2(enc_h1)))
        enc_h3 = F.relu(self.BN3(self.enc_3(enc_h2)))

        z1 = self.BN4(self.z1_layer(enc_h3))
        z2 = self.BN5(self.z2_layer(z1))
        z3 = self.BN6(self.z3_layer(z2))

        dec_h1 = F.relu(self.BN7(self.dec_1(z3)))
        dec_h2 = F.relu(self.BN8(self.dec_2(dec_h1)))
        dec_h3 = F.relu(self.BN9(self.dec_3(dec_h2)))
        x_bar = self.x_bar_layer(dec_h3)

        return x_bar, enc_h1, enc_h2, enc_h3, z3, z2, z1, dec_h3


class RAdam(Optimizer):
    """RAdam optimizer class.

    Parameters
    ----------
    params :
        model parameters.
    lr : float optional
        learning rate.
    betas : tuple optional
        coefficients used for computing running averages of gradient and its square.
    eps : float optional
        term added to the denominator to improve numerical stability.
    weight decay : float optional
        weight decay (L2 penalty).
    degenerated_to_sgd : bool optional
        degenerated to SGD or not.

    """

    def __init__(self, params, lr=1e-3, betas=(0.9, 0.999), eps=1e-8, weight_decay=0, degenerated_to_sgd=True):
        if not 0.0 <= lr:
            raise ValueError("Invalid learning rate: {}".format(lr))
        if not 0.0 <= eps:
            raise ValueError("Invalid epsilon value: {}".format(eps))
        if not 0.0 <= betas[0] < 1.0:
            raise ValueError("Invalid beta parameter at index 0: {}".format(betas[0]))
        if not 0.0 <= betas[1] < 1.0:
            raise ValueError("Invalid beta parameter at index 1: {}".format(betas[1]))

        self.degenerated_to_sgd = degenerated_to_sgd
        if isinstance(params, (list, tuple)) and len(params) > 0 and isinstance(params[0], dict):
            for param in params:
                if 'betas' in param and (param['betas'][0] != betas[0] or param['betas'][1] != betas[1]):
                    param['buffer'] = [[None, None, None] for _ in range(10)]
        defaults = dict(lr=lr, betas=betas, eps=eps, weight_decay=weight_decay,
                        buffer=[[None, None, None] for _ in range(10)])
        super(RAdam, self).__init__(params, defaults)

    def __setstate__(self, state):
        super(RAdam, self).__setstate__(state)

    def step(self, closure=None):

        loss = None
        if closure is not None:
            loss = closure()

        for group in self.param_groups:

            for p in group['params']:
                if p.grad is None:
                    continue
                grad = p.grad.data.float()
                if grad.is_sparse:
                    raise RuntimeError('RAdam does not support sparse gradients')

                p_data_fp32 = p.data.float()

                state = self.state[p]

                if len(state) == 0:
                    state['step'] = 0
                    state['exp_avg'] = torch.zeros_like(p_data_fp32)
                    state['exp_avg_sq'] = torch.zeros_like(p_data_fp32)
                else:
                    state['exp_avg'] = state['exp_avg'].type_as(p_data_fp32)
                    state['exp_avg_sq'] = state['exp_avg_sq'].type_as(p_data_fp32)

                exp_avg, exp_avg_sq = state['exp_avg'], state['exp_avg_sq']
                beta1, beta2 = group['betas']

                exp_avg_sq.mul_(beta2).addcmul_(1 - beta2, grad, grad)
                exp_avg.mul_(beta1).add_(1 - beta1, grad)

                state['step'] += 1
                buffered = group['buffer'][int(state['step'] % 10)]
                if state['step'] == buffered[0]:
                    N_sma, step_size = buffered[1], buffered[2]
                else:
                    buffered[0] = state['step']
                    beta2_t = beta2 ** state['step']
                    N_sma_max = 2 / (1 - beta2) - 1
                    N_sma = N_sma_max - 2 * state['step'] * beta2_t / (1 - beta2_t)
                    buffered[1] = N_sma

                    # more conservative since it's an approximated value
                    if N_sma >= 5:
                        step_size = math.sqrt(
                            (1 - beta2_t) * (N_sma - 4) / (N_sma_max - 4) * (N_sma - 2) / N_sma * N_sma_max / (
                                    N_sma_max - 2)) / (1 - beta1 ** state['step'])
                    elif self.degenerated_to_sgd:
                        step_size = 1.0 / (1 - beta1 ** state['step'])
                    else:
                        step_size = -1
                    buffered[2] = step_size

                # more conservative since it's an approximated value
                if N_sma >= 5:
                    if group['weight_decay'] != 0:
                        p_data_fp32.add_(-group['weight_decay'] * group['lr'], p_data_fp32)
                    denom = exp_avg_sq.sqrt().add_(group['eps'])
                    p_data_fp32.addcdiv_(-step_size * group['lr'], exp_avg, denom)
                    p.data.copy_(p_data_fp32)
                elif step_size > 0:
                    if group['weight_decay'] != 0:
                        p_data_fp32.add_(-group['weight_decay'] * group['lr'], p_data_fp32)
                    p_data_fp32.add_(-step_size * group['lr'], exp_avg)
                    p.data.copy_(p_data_fp32)

        return loss


class MeanAct(nn.Module):
    """Mean activation class.
    """

    def __init__(self):
        super(MeanAct, self).__init__()

    def forward(self, x):
        return torch.clamp(torch.exp(x), min=1e-5, max=1e6)


class DispAct(nn.Module):
    """Dispersion activation class.
    """

    def __init__(self):
        super(DispAct, self).__init__()

    def forward(self, x):
        return torch.clamp(F.softplus(x), min=1e-4, max=1e4)

